/*****************************************************************************/
/*                                                                           */
/*  QaSEngine                                                                */
/*                                                                           */
/*                                                                           */
/*     Description:                                                          */
/*             The QaS infinite loop that dispatches QoS REQ and ACK         */
/*             messages as well as timer messages.                           */
/*                                                                           */
/*     History:                                                              */
/*             23/10/2002 - Fabien Chanussot - Creation                      */
/*                                                                           */
/*****************************************************************************/

//###FC - TBD package QaS;

/*****************************************************************************/
/* Use                                                                       */
/*****************************************************************************/

import java.util.*;
import java.io.*;

import org.xml.sax.InputSource;

/*****************************************************************************/
/* QaSEngine class:                                                          */
/*                  Infinite loop that keeps the system alive.               */
/*****************************************************************************/
class QaSEngine
{
    /*************************************************************************/
    /* Private                                                               */
    /*************************************************************************/
    
    QaSMapper myQaSMapper;       //QoS mapping algorithm
    QaSXMLIf  myQaSXMLIf;        //Interface with XML parser
    Timer myTimer;               //...
    
    static final String QAS_CONF_DEFAULT_FILE = "/.QoSINUS/dynamicConf.xml";
    
    
    class qasPeriodicThread extends TimerTask //Nested class
    {
        public void run() //Periodically refresh the soft state
	{
	    myQaSMapper.heartBeat();
        }
    }
    
    
    /*************************************************************************/
    /* Public                                                                */
    /*************************************************************************/
    
    
    /*
     * start(): create a thread that enters the message loop
     */
    public void start(int period) throws IOException
    {
	FileInputStream aStream;
	InputSource aSource;
	QaSXMLDoc aXMLConfDoc;
	
	System.out.println("### Started QasEngine");
	
	//Create the XML parser
	myQaSXMLIf = new QaSXMLIf();
	
	//Parse the configuration XML document
	aStream = new FileInputStream(System.getProperty("user.home") + QAS_CONF_DEFAULT_FILE);
	aSource = new InputSource(aStream);
	aXMLConfDoc = myQaSXMLIf.parse(aSource);
	
	//Create the mapper
	myQaSMapper = new QaSDynamicMapper1(aXMLConfDoc); //###FC - TMP: pick the sub-type dynamically
	
	//Start a timer
	myTimer = new Timer(); //###FC - TBD: check if a timer is not already running...
	
	//Launch the corresponding periodic thread
        myTimer.schedule(new qasPeriodicThread(), period*1000, period*1000);
    }
    
    
    /*
     * stop(): stop the message loop thread
     */
    public void stop() 
    {
	//Stop the timer
	myTimer.cancel();
    }
    
    
    /*
     * post(): send a message to the message loop thread's queue
     */
    public byte[] post(byte[] message, int type)
    {
	String aString;
	StringBufferInputStream aStream;
	InputSource aSource;
	QaSXMLDoc aDocument;
	String aMessageType;
	byte[] anAck = {1};
	
	
	aString = new String(message, 0, message.length);
	
	System.out.println("### Received: " + aString); //###FC - DEBUG
	
	//Parse the SLS
	aStream = new StringBufferInputStream(aString);
	aSource = new InputSource(aStream);

	aDocument = myQaSXMLIf.parse(aSource);
	
	//Switch on message type
	aMessageType = aDocument.getAttribute("type");
	
	if ( aMessageType.compareTo("QOS_REQ") == 0 )
	    { //QoSINUS reservation request
		//Send SLS to the mapper
		myQaSMapper.reserve(aDocument);
		
		//Send answer to client
		return(myQaSXMLIf.serialize(aDocument));
	    }
	else if ( aMessageType.compareTo("QOS_RST") == 0 )
	    { //QoSINUS release request
	      System.out.println("###About to QOS_RST...\n");
		myQaSMapper.release(aDocument);

		//Send ACK to client
		return(myQaSXMLIf.serialize(aDocument));
	    }
	else
	    { //Unknown message type
		System.err.println("Error unknown message type:\n" + aString);
		return(null);
	    }
    }
    
    
    /*
     * observeHTML(): built an HTML status page
     */
    public void observeHTML(StringBuffer pString)
    {
	pString.append("<table border=\"0\" style=\"width: 100%;\"><tbody><tr align=\"center\" style=\"font-family: helvetica,arial,sans-serif;\"><td><h1 style=\"background-color: rgb(0, 0, 255); color: rgb(255, 255, 255);\">QoSINUS Active Service</h1></td></tr></tbody></table>");
	
	myQaSMapper.observeHTML(pString);
    }
}
