/*****************************************************************************/
/*                                                                           */
/*  QaSExecEnvIf                                                             */
/*                                                                           */
/*                                                                           */
/*     Description:                                                          */
/*             Classes that implement the QaS entry point. This can be       */
/*             either a standalone Java application or a Tamanoir service.   */
/*                                                                           */
/*     History:                                                              */
/*             22/10/2002 - Fabien Chanussot - Creation                      */
/*                                                                           */
/*****************************************************************************/

//###FC - TBD package QaS;


/*****************************************************************************/
/* Use                                                                       */
/*****************************************************************************/

import java.util.*;
import java.net.*;
import java.io.*;


/*****************************************************************************/
/* QaSExecEnvIf class:                                                       */
/*                     System entry point. Virtual.                          */
/*****************************************************************************/
abstract class QaSExecEnvIf
{
    /*************************************************************************/
    /* Private                                                               */
    /*************************************************************************/
    
    /*************************************************************************/
    /* Public                                                                */
    /*************************************************************************/
    
    static QaSEngine myQaSEngine; //Start and stopable infinite loop
    
    
    /*
     * getPayload(): extract sls from data received on socket
     */
    public byte[] getPayload(DataInputStream pInputStream) throws IOException    
    {
	byte someBytes[] = new byte[0];
	
	System.out.println( "### Called getPayload" );	
	return someBytes;
    }
    
    
    /*
     * sendPayload(): send sls thru socket
     */
    public void sendPayload(DataOutputStream pOutputStream, byte[] pPayload) throws IOException
    {
	System.out.println( "### Called sendPayload" );	
    }

}


/*****************************************************************************/
/* QaSConnection class:                                                      */
/*                      Class that handle a connection with a client         */
/*****************************************************************************/
class QaSConnection extends Thread
{
    /*************************************************************************/
    /* Protected                                                             */
    /*************************************************************************/
    
    protected Socket client;
    protected DataInputStream in;
    protected DataOutputStream out;
    protected QaSExecEnvIf execEnv;
    
    
    /*************************************************************************/
    /* Public                                                                */
    /*************************************************************************/
    
    /*
     * Constructors
     */
    public QaSConnection(Socket clientSocket, QaSExecEnvIf anExecEnv)
    {
	this.client = clientSocket;
	this.execEnv = anExecEnv;
	
	try
	    {
		in = new DataInputStream(client.getInputStream());
		out = new DataOutputStream(client.getOutputStream());
	    }
	catch (IOException e)
	    {
		try {this.client.close();} catch (IOException e2){}
		System.err.println("Error: exception while geting socket streams: " + e);
	    }
	
	/* A new connection arrived; handle it */
	this.start();
    }
    
    
    /*
     * run(): post the SLSin and send the SLSout back to the client
     */
    public void run()
    {
	byte[] slsOut;
	byte[] aPayload;
	String aString;
	
	try
	    {
		//Extract payload and post it to QaSEngine
		aPayload = this.execEnv.getPayload(this.in);
		aString = new String(aPayload, 0, aPayload.length);
		System.out.println("Received: " + aPayload + "\n");
		
		if ( aPayload.length !=0 )
		    {
			//Mutex on myQaSEngine
			synchronized (this.execEnv.myQaSEngine)
			    {
				slsOut = this.execEnv.myQaSEngine.post(aPayload, 0);
			    }
			
			//Send response back to the client
			aString = new String(slsOut, 0, slsOut.length);
			System.out.println("Replying: " + aString + "\n");
			
			this.execEnv.sendPayload(this.out, slsOut);
		    }
	    }
	catch (IOException e){}
	//###FC - TBDecided Close the connection at some point???
    }
}


/*****************************************************************************/
/* QaSTCPServer class:                                                       */
/*                     TCP server that handles connection requests           */
/*****************************************************************************/
class QaSTCPServer extends Thread
{
    /*************************************************************************/
    /* Protected                                                             */
    /*************************************************************************/
    
    protected int port;
    protected ServerSocket listenSocket;
    protected QaSExecEnvIf execEnv;
    
    
    /*************************************************************************/
    /* Public                                                                */
    /*************************************************************************/

    /*
     * fail(): deal with errors
     */
    public void fail(Exception e, String message)
    {
	System.err.println(message + ": " + e);
	System.exit(1);
    }
    
    
    /*
     * Constructors
     */
    public QaSTCPServer(int port, QaSExecEnvIf pExecEnv)
    {
	/* Initialization */
	this.port = port;
	this.execEnv = pExecEnv;
	
	/* Create socket server */
	try
	    {
		this.listenSocket = new ServerSocket(port);
	    }
	catch (IOException e)
	    {
		fail(e, "Error: exception creating server socket");
	    }
	
	/* Start the server */
	//###FC - TBD: clean clean-up
	this.start();
    }
    
    
    /*
     * run(): loop listening for and accepting new connections
     */
    public void run()
    {
	try
	    {
		while (true)
		    {
			/* Listen for incoming connections */
			Socket clientSocket = listenSocket.accept();
			
			/* Handle communication */
			QaSConnection aConnection = new QaSConnection(clientSocket, execEnv);
		    }
	    }
	catch (IOException e)
	    {
		fail(e, "Error: exception while listening");
	    }
    }
}


/*****************************************************************************/
/* QaSStandalone class:                                                      */
/*                      System entry point. Standalone version.              */
/*****************************************************************************/
class QaSStandalone extends QaSExecEnvIf
{
    /*************************************************************************/
    /* Private                                                               */
    /*************************************************************************/
    
    int port; //Port to listen for active packets
    static final int ANEP_VERSION = 1; 
    static final int TAMANOIR_ANEP_TYPE = 20;
    static final int ANEP_LENGTH_WO_TLV = 8;
    QaSTCPServer myQaSTCPServer; //TCP server the handles TCP connections
    
    
    /*************************************************************************/
    /* Public                                                                */
    /*************************************************************************/
    
    /*
     * Constructors
     */
    public QaSStandalone(int port)
    {
	this.port = port;
	
	/* Create and start the QaS engine */
	myQaSEngine = new QaSEngine();
	
	try
	    {
		myQaSEngine.start(1);
	    }
	catch (IOException e)
	    {
		System.err.println("Error starting engine: " + e);
		return;
	    }
	
	/* Create TCP server */
	myQaSTCPServer = new QaSTCPServer(port, this);
    }
    
    
    /*
     * Parse and throw the ANEP header away
     */
    public byte[] getPayload(DataInputStream pInputStream) throws IOException
    {
	int dummyInt;
	int hdrLen, pktLen;
	byte[] dummyByte, payload;
	
	
	//Throw header away
        dummyInt = (int) Marshalling.getByte(pInputStream); //version
	if ( dummyInt != ANEP_VERSION ) 
	  { //This is not an ANEP packet
	    return new byte[0];
	  }
	
        dummyInt = (int) Marshalling.getByte(pInputStream); //flags
        dummyInt = (int) Marshalling.getByte(pInputStream); //typeid
        dummyInt = (int) Marshalling.getByte(pInputStream);
	
        hdrLen = (int)Marshalling.getByte(pInputStream) & 0xff;
        hdrLen = hdrLen * 256 + ((int)Marshalling.getByte(pInputStream) & 0xff);
	
        pktLen = (int)Marshalling.getByte(pInputStream) & 0xff;
        pktLen = pktLen * 256 + ((int)Marshalling.getByte(pInputStream) & 0xff);
	
	dummyByte = new byte[hdrLen - 8];
	dummyInt = pInputStream.read(dummyByte, 0, hdrLen - 8);
	
	//Read the payload
	payload = new byte[pktLen - hdrLen];
	dummyInt = pInputStream.read(payload, 0, pktLen - hdrLen);
	
	
	//Done !
        return payload;
    }
    
    
    /*
     * sendPayload(): build a minimal ANEP header and send it together with the payload
     */
    public void sendPayload(DataOutputStream pOutputStream, byte[] pPayload) throws IOException
    {
	byte[] anAnep = new byte[8];
	int hdrlen = ANEP_LENGTH_WO_TLV;
	int pktLen;
	
	
	//Build and send a minimalist ANEP header
	Marshalling.putByte(pOutputStream, (byte)ANEP_VERSION);//version
	Marshalling.putByte(pOutputStream, (byte) 0);//flags

	Marshalling.putByte(pOutputStream, (byte) (TAMANOIR_ANEP_TYPE / 256));//typeid
	Marshalling.putByte(pOutputStream, (byte) (TAMANOIR_ANEP_TYPE % 256));
	
	Marshalling.putByte(pOutputStream, (byte) (hdrlen / 256));//header length
	Marshalling.putByte(pOutputStream, (byte) (hdrlen % 256));
	
	pktLen = hdrlen + pPayload.length;
	
	Marshalling.putByte(pOutputStream, (byte) (pktLen / 256));//total length
	Marshalling.putByte(pOutputStream, (byte) (pktLen % 256));
	
	//Send the poayload
	pOutputStream.write(pPayload, 0, pPayload.length);
	
	//###FC - TBD
	System.out.println(getInfo());	
    }
    
    /*
     * getInfo(): build HTML status page
     */
    public String getInfo()
    {
	StringBuffer aStringBuff = new StringBuffer();
	
	//Header
	aStringBuff.append("<html><head><meta http-equiv=\"refresh\" content=\"3\"><title>QaS Obs</title><link rel=\"stylesheet\" type=\"text/css\" href=\"css/main.css\"></head><body bgcolor=white><table border=0 width=100%><tr><td nowrap>");
	
	//Body
	synchronized (myQaSEngine)
	    {
		myQaSEngine.observeHTML(aStringBuff);
	    }
	
	//Footer
	aStringBuff.append("</td></tr></table></body></html>");	
	
	return(new String(aStringBuff));
    }
    
    
    /*
     * main(): guess what...
     */
    public static void main  (String[] args)
    {
	QaSStandalone myQaSStandalone;
	int i;
	int port = 1025;
	
	
	/* Parse arguments */
	for (i = 0; i < args.length; i++)
	    {
		if ( args[i].equals ("--help") || args[i].equals ("-h") )
		    {
			System.out.println("Usage: QaSStandalone [-p port]");
			System.exit(1);
		    }
		else if ( args[i].equals ("-p") )
		    {
			i++;
			if ( i >= args.length )
			    {
				System.out.println("Usage: QaSStandalone [-p port]");
				System.exit(1);
			    }
			port = Integer.parseInt(args[i]);
		    }
		else
		    {
			System.out.println("Usage: QaSStandalone [-p port]");
			System.exit(1);
		    }
	    }
	
	/* Create exec env object */
	myQaSStandalone= new QaSStandalone(port);
    } 
} // class QaSStandalone


/*****************************************************************************/
/* QaS class:                                                                */
/*                   System entry point. TAN Service version.                */
/*****************************************************************************/
class QaS extends Service
{
    final static boolean DBG = false;
    static QaSEngine myQaSEngine; //Start and stopable infinite loop
    
    
    QaS() 
    {
	if(DBG) System.out.println( "QaS : build" );
    }
    
    
    public void run()
    {
	byte[] slsOut;
	
	name = "QaS";
	institution = "RESO";
	author = "FC";
	date = "2003";
	comment = "QoSINUS active service";
	serverbroker = "localhost";
	
	/* Launch QaSEngine if it's the first service invocation */
	if ( myQaSEngine == null )
	    {
		myQaSEngine = new QaSEngine();
		
		try
		    {
			myQaSEngine.start(1);
		    }
		catch (IOException e)
		    {
			System.err.println("Error starting engine: " + e);
			return;
		    }
	    }
	
	/* Post SLSin to QaSEngine */
	//Mutex on myQaSEngine
	synchronized (myQaSEngine)
	    {
		slsOut = myQaSEngine.post(net.getANEPkt().payload, 0);
	    }
	
	/* Send SLSout back to client */
	net.makeANEPkt(net.serviceNameinBytes, 
		       net.destId.getBytes(),
		       net.srcId.getBytes(),
		       net.destId.getBytes(),
		       net.proxyTANId.getBytes(),
		       slsOut);
	try
	    {
		net.sendANEPkt();
	    }
	catch (IOException e)
	    {
		System.err.println("Error sending SLSout: " + e);
	    }
	
	//###FC - TBChecked: do we need to quit the first object running so that the QaSEngine is not killed?
	//###FC - TBD: when Tamanoir is excited, kill QaSEngine
    }

    /*
     * getInfo(): build HTML status page
     */
    public String getInfo()
    {
	StringBuffer aStringBuff = new StringBuffer();
	
	//Header
	aStringBuff.append("<html><head><meta http-equiv=\"refresh\" content=\"3\"><title>QaS Obs</title></head><body bgcolor=white><table border=0 width=100%><tr><td nowrap>");
	
	//Body
	synchronized (myQaSEngine)
	    {
		myQaSEngine.observeHTML(aStringBuff);
	    }
	
	//Footer
	aStringBuff.append("</body></html>");	
	
	return(new String(aStringBuff));
    }

}

