/*****************************************************************************/
/*                                                                           */
/*  QaSMapper                                                                */
/*                                                                           */
/*                                                                           */
/*     Description:                                                          */
/*             The QoS classes mapping algorithm.                            */
/*                                                                           */
/*     History:                                                              */
/*             24/01/2003 - Fabien Chanussot - Creation                      */
/*                                                                           */
/*****************************************************************************/

//###FC - TBD package QaS;

/*****************************************************************************/
/* Use                                                                       */
/*****************************************************************************/

import java.util.*;
import java.io.*;


/*****************************************************************************/
/* QaSPickParam class:                                                       */
/*                                                                           */
/*****************************************************************************/
class QaSPickParam
{
    /*************************************************************************/
    /* Public                                                                */
    /*************************************************************************/
    
    int dscp;
    
    
    /*
     * Constructor
     */
    public QaSPickParam(int aDscp)
    {
	this.dscp = aDscp;
    }
}


/*****************************************************************************/
/* QaSMapper class:                                                          */
/*                     The QoS classes mapping algorithm. Virtual.           */
/*****************************************************************************/
abstract class QaSMapper
{
    /*************************************************************************/
    /* Protected ###FC - TBDone                                              */
    /*************************************************************************/
    
    //Information about data streams
    QaSStreamStatus myQaSStreamStatus;
    
    //Information about DiffServ classes
    QaSClassStatus myQaSClassStatus;
    
    //Configuration data
    QaSXMLDoc myXMLConfDoc;
    
    
    /*************************************************************************/
    /* Public                                                                */
    /*************************************************************************/
    
    /*
     * Constructors
     */
    public QaSMapper(QaSXMLDoc pXMLConfDoc)
    {
	myXMLConfDoc = pXMLConfDoc;
	
	//Create Stream Status table
	myQaSStreamStatus = new QaSStreamStatus();
	
	//Create Class Status table
	myQaSClassStatus = new QaSClassStatus();
	
	//Initialization hook
	initHook();
    }
    
    
    /*
     * initHook(): initialisation hook
     */
    public void initHook()
    {
	//###FC - TBD: virtual
	System.out.println("### Called QaSMapper.initHook()");
    }
    
    
    /*
     * acceptHook(): acceptability hook
     */
    public QaSPickParam acceptHook(QaSXMLDoc pDocument)
    {
	//###FC - TBD: virtual
	System.out.println("### Called QaSMapper.acceptHook()");
	
	return(null);
    }
    
    
    /*
     * reserveHook(): resource reservation hook
     */
    public void reserveHook(QaSXMLDoc pDocument, int pLocal, int pChosenClass)
    {
	//###FC - TBD: virtual
	System.out.println("### Called QaSMapper.reserveHook()");
    }
    
    
    /*
     * releaseHook(): resource reservation hook
     */
    public void releaseHook(int pDscp)
    {
	//###FC - TBD: virtual
	System.out.println("### Called QaSMapper.releaseHook()");
    }
    
    
    /*
     * heartBeatHook(): pulse hook
     */
    public void heartBeatHook()
    {
    }
    
    
    /*
     * heartBeat(): called on timer's tick
     */
    public void heartBeat()
    {
	int i;
	
	
	//###FC - DEBUG System.out.print(".");
	
	//Call class specific hook
	heartBeatHook();

	//Traverse the stream table
	for ( i = 0; i < myQaSStreamStatus.QAS_MAX_NUM_STREAMS; i++ )
	    {
		boolean timedout = false;
		QaSKernelParam param;
		
		
		//Ask kernel module about streams that have been active since the last pulse
		param = new QaSKernelParam();
		param.local = i;
		QaSKernelIf.getNetfilter(param);
		
		if ( param.diffserv != i )
		  System.out.println("Stream " + i + " DSCP " + param.diffserv);
		
		//If the stream is active, rearm timeout counter
		if ( param.activity > 0 )
		    {
			myQaSStreamStatus.refresh(i);
			//System.out.println("Rearming counter for " + i);
		    }
		//else, decrement timeout counters
		else
		    timedout = myQaSStreamStatus.pulse(i); //###FC - TBChecked: do not decrement streams that are reserved
		
		//Reset activity flag
		param.activity = 0;
		QaSKernelIf.setNetfilter(param);
		
		//Release ressources that timed out
		if ( timedout )
		    releaseDSCP(i);
	    }
	
	System.out.println("");
	
	myQaSStreamStatus.dump();
	//myQaSClassStatus.dump();	
    }
    
    
    /*
     * reserve(): called on SLS reception
     */
    public QaSXMLDoc reserve(QaSXMLDoc pDocument)
    {
	int aLocalDSCP = 0;
	int aDiffServDSCP;
	QaSPickParam pickParam;
	int ret;
	QaSKernelParam param = new QaSKernelParam();
	int aTimeout;
	String aString;
	String aClassName;
	String aCommand;
	
	
	System.out.println("### Called QaSMapper.reserve()");
	
	//Decide if the request is acceptable
	pickParam = acceptHook(pDocument);
	aDiffServDSCP = pickParam.dscp;
	
	// if ( aDiffServDSCP != 0 ) //###FC - TBD: do that based on the dynamic type
	  {
	      //Add the stream in the stream status table
	      aClassName = myXMLConfDoc.getFilteredAttribute("configured/class", "dscp", Integer.toString(aDiffServDSCP), "name");
	      aString = myXMLConfDoc.getFilteredAttribute("configured/class", "dscp", Integer.toString(aDiffServDSCP), "timeout");
	      aTimeout = Integer.parseInt(aString);
	      
	      aLocalDSCP = myQaSStreamStatus.addStream(pDocument, aDiffServDSCP, aClassName, aTimeout);
	      
	      if ( aLocalDSCP < 0 )
		  {
		      System.out.println("QoSINUS! Too many stream in parallel!\n");
		  }
	      
	      //Reserve the resources
	      if ( aLocalDSCP >= 0 )
		  {
		      reserveHook(pDocument, aLocalDSCP, aDiffServDSCP);
		      
		      //Configure kernel module
		      param.local = aLocalDSCP;
		      param.diffserv = aDiffServDSCP; //###FC - TBD: get rate, burst and backlog
		      param.bucketrate = myQaSStreamStatus.myStreamTable[aLocalDSCP].myRate;
		      ret = QaSKernelIf.setNetfilter(param);

		      System.out.println("### Set DSCP " + param.diffserv + " for stream " + param.local);
		      
		      aCommand = myXMLConfDoc.getAttribute("token_bucket/script");
		      ret = QaSKernelIf.setTc(param, aCommand);
		  }
	      else
		  {
		      aLocalDSCP = 0;
		  }
	  }
	
	//Compose response
	pDocument.setAttributeCreate("flowid/dscp", Integer.toString(aLocalDSCP));
	//###FC - DEBUG ONLY pDocument.setAttributeCreate("flowid/dscp", Integer.toString(aDiffServDSCP));
	pDocument.setAttribute("type", "QOS_ACK");
	
	
	//Done!
	return(pDocument);
    }
    
    
    /*
     * release(): called when when the QoS session is closed
     */
    public QaSXMLDoc release(QaSXMLDoc pDocument)
    {
	int aLocalDSCP;
	
	
	//Get the stream Id
	aLocalDSCP = Integer.parseInt(pDocument.getAttribute("flowid/dscp"));
	
	System.out.println("### Called QaSMapper.release of localId " + aLocalDSCP);
	
	//Release the corresponding stream
	releaseDSCP(aLocalDSCP);
	
	//Compose response
	pDocument.setAttribute("type", "QOS_ACK");
	
	
	//Done!
	return(pDocument);
    }
    
    
    /*
     * releaseDSCP(): release ressources attached to a stream ID
     */
    public void releaseDSCP(int pDscp) 
    {
	QaSKernelParam param = new QaSKernelParam();
	int ret;
	String aCommand;
	
	
	System.out.println("### Called QaSMapper.released()");
	
	//Let the kernel module know
	param.local = pDscp;
	param.diffserv = pDscp;
	param.bucketrate = Integer.parseInt(myXMLConfDoc.getAttribute("configured/link_bandwidth/value")); //In Mbps
	ret = QaSKernelIf.setNetfilter(param);
	aCommand = myXMLConfDoc.getAttribute("token_bucket/script");
	ret = QaSKernelIf.setTc(param, aCommand); //###FC - TBD: get burst and backlog
	
	//Release the ressources
	releaseHook(pDscp);
	
	//Remove entry from the stream status table
	myQaSStreamStatus.removeStream(pDscp);
    }
    
    
    /*
     * observeHTML(): built an HTML status page
     */
    public void observeHTML(StringBuffer pString)
    {
	//###FC - TBD: virtual
    }
}


/*****************************************************************************/
/* QaSStaticMapper class:                                                    */
/*                         Simple static mapping algorithm.                  */
/*****************************************************************************/
class QaSStaticMapper extends QaSMapper
{
    static final int QAS_EF = 46; //###FC - TBD: this should be read from conf file
    static final int QAS_AFTCP = 47;
    static final int QAS_AFUDP = 48;
    
    
    /*************************************************************************/
    /* Private                                                               */
    /*************************************************************************/
    
    /*
     * mapQualitativeValues(): Translate qualitative guarantee requests into quantitative ones
     */
    void mapQualitativeValues(QaSXMLDoc pDocument)
    {
	//Replace qualitative values with quantitative ones
	pDocument.replaceAttributeByValue("guarantees/delay/*", "high", myXMLConfDoc.getAttribute("qualitative_mapping/delay/high"));
	pDocument.replaceAttributeByValue("guarantees/delay/*", "medium", myXMLConfDoc.getAttribute("qualitative_mapping/delay/medium"));
	pDocument.replaceAttributeByValue("guarantees/delay/*", "low", myXMLConfDoc.getAttribute("qualitative_mapping/delay/low"));
	
	pDocument.replaceAttributeByValue("guarantees/rate/*", "high", myXMLConfDoc.getAttribute("qualitative_mapping/rate/high"));
	pDocument.replaceAttributeByValue("guarantees/rate/*", "medium", myXMLConfDoc.getAttribute("qualitative_mapping/rate/medium"));
	pDocument.replaceAttributeByValue("guarantees/rate/*", "low", myXMLConfDoc.getAttribute("qualitative_mapping/rate/low"));
	
	pDocument.replaceAttributeByValue("guarantees/loss/*", "high", myXMLConfDoc.getAttribute("qualitative_mapping/loss/high"));
	pDocument.replaceAttributeByValue("guarantees/loss/*", "medium", myXMLConfDoc.getAttribute("qualitative_mapping/loss/medium"));
	pDocument.replaceAttributeByValue("guarantees/loss/*", "low", myXMLConfDoc.getAttribute("qualitative_mapping/loss/low"));
    }
    
    
    /*
     * pickDiffServClass(): pick the best DiffServ class corresponding to the request
     */
    QaSPickParam pickDiffServClass(QaSXMLDoc pDocument)
    {
	int diffServClass = 0;
	int delayReq;
	int rateReq;
	String aString;
	QaSPickParam aParam;
	
	
	//Delay is taken into account first
	aString = pDocument.getAttribute("guarantees/delay/value");

	//if ( aString.length() == 0 )
	//    { //A range is specified
	//	aString = pDocument.getAttribute("guarantees/delay/range/max"); //###FC - TBD: take min value into account
	//    }
	
	
	//If small delay is requested, use EF
	if ( aString.length() != 0 )
	    {
		int efLimit;
		
		efLimit = Integer.parseInt(myXMLConfDoc.getAttribute("thresholds/EFDelayLimit/value")); //###FC - TBD: error handling
		delayReq = Integer.parseInt(aString); //###FC - TBD: error checking (catch exception)
		
		if ( delayReq <= efLimit )
		    {
			diffServClass = QAS_EF;
		    }
	    }
	
	//Get requested rate
	aString = pDocument.getAttribute("guarantees/rate/value");
	
	if ( aString.length() != 0 )
	    {
		rateReq = Integer.parseInt(aString); //###FC - TBD: error checking (catch exception)
		
		//If no diffServ has been picked yet, choose AF
		if ( diffServClass == 0 )
		    if (  pDocument.getAttribute("flowId/proto/name").compareTo("tcp") == 0 )
			{
			    diffServClass = QAS_AFTCP;
			}
		    else
			{
			    diffServClass = QAS_AFUDP;
			}
		
	    }
	else
	    //Use default rate
	    rateReq = Integer.parseInt(myXMLConfDoc.getAttribute("defaults/rate/value")); //###FC - TBD: error handling
	
	//Check if requested rate is available, and declass when needed
	if ( diffServClass == QAS_EF )
	    if ( ( myQaSClassStatus.myClassTable[diffServClass].configured - myQaSClassStatus.myClassTable[diffServClass].allocated ) < rateReq )
		{
		    if (  pDocument.getAttribute("flowId/proto/name").compareTo("tcp") == 0 )
			{
			    diffServClass = QAS_AFTCP;
			}
		    else
			{
			    diffServClass = QAS_AFUDP;
			}
		}
	
	if ( ( diffServClass == QAS_AFTCP ) || ( diffServClass == QAS_AFUDP ) )
	    if ( ( myQaSClassStatus.myClassTable[diffServClass].configured - myQaSClassStatus.myClassTable[diffServClass].allocated ) < rateReq )
		{
		    diffServClass = 0;
		    System.out.println("### Pick 0.......");    
		}

	aParam = new QaSPickParam(diffServClass);
	
	//Done!
	return(aParam);
    }
    
    
    /*************************************************************************/
    /* Public                                                                */
    /*************************************************************************/
    
    /*
     * Constructors
     */
    public QaSStaticMapper(QaSXMLDoc pXMLConfDoc)
    {
	super(pXMLConfDoc);
    }
    
    
    /*
     * initHook(): initialisation hook
     */
    public void initHook()
    {	
	int ret;
	int totalBW;
	int i;
	String aCommand;
	
        //System.out.println("### Called QaSStaticMapper.initHook()");
	
	//Initialize the classes configured bandwidth static values
	totalBW = Integer.parseInt(myXMLConfDoc.getAttribute("configured/link_bandwidth/value")); //###FC - TBD: error handling
	
        System.out.println("    Configured BW: total :" + totalBW);
	
	for ( i = 0; i < QaSClassStatus.QAS_MAX_NUM_CLASSES; i++ )
	    {
		String aString;
		String aClassName;
		
		aString = myXMLConfDoc.getFilteredAttribute("configured/class", "dscp", Integer.toString(i), "bandwidth");
		aClassName = myXMLConfDoc.getFilteredAttribute("configured/class", "dscp", Integer.toString(i), "name");
		
		if ( aString.compareTo("") != 0 )
		    {
			//Configure class
			myQaSClassStatus.myClassTable[i].configured = ( Integer.parseInt(aString) * totalBW ) / 100;
			
			//Reserve DSCP to allow QoSINUS override
			myQaSStreamStatus.forceStream(i, false, null, i, aClassName, 0, 0, 0);
			
			//###FC - TBD System.out.println("    classe " + i + ": " + myQaSClassStatus.myClassTable[i].configured);
		    }
		
		//Just clean up any leftover module configuration 
		QaSKernelParam param = new QaSKernelParam();
		param.local = i;
		param.diffserv = i;
		param.bucketrate = Integer.parseInt(myXMLConfDoc.getAttribute("configured/link_bandwidth/value")); //In Mbps
		ret = QaSKernelIf.setNetfilter(param);
		aCommand = myXMLConfDoc.getAttribute("token_bucket/script");
		ret = QaSKernelIf.setTc(param, aCommand); //###FC - TBD: get burst and backlog
	    }
    }
    
    
    /*
     * acceptHook(): acceptability hook
     */
    public QaSPickParam acceptHook(QaSXMLDoc pDocument)
    {
	int diffServDSCP;
	QaSPickParam aParam;
	
	
	//System.out.println("### Called QaSStaticMapper.acceptHook()");
	
	//Map qualitative values into quantitative ones
	mapQualitativeValues(pDocument);
	
	//Policy checking could come here.
	
	//Pick the best DiffServ class
	aParam = pickDiffServClass(pDocument);
	
	System.out.println("Picked DiffServ=" + aParam.dscp);
	
	return(aParam);
    }
    
    /*
     * reserveHook(): resource reservation hook
     */
    public void reserveHook(QaSXMLDoc pDocument, int pLocal, int pChosenClass)
    {
	String aString;
	int rateReq;
	
	
	//System.out.println("### Called QaSStaticMapper.reserveHook()");
	
	//Get the rate requested
	aString = pDocument.getAttribute("guarantees/rate/value");
	
	if ( aString.length() != 0 )
	    rateReq = Integer.parseInt(aString); //###FC - TBD: error checking (catch exception)
	else
	    //Use default rate
	    rateReq = Integer.parseInt(myXMLConfDoc.getAttribute("defaults/rate/value")); //###FC - TBD: error handling
	
	//Update class status
	myQaSClassStatus.myClassTable[pChosenClass].allocated += rateReq;
	
	//Update stream status
	myQaSStreamStatus.myStreamTable[pLocal].myRate = rateReq;
    }
    
    
    /*
     * releaseHook(): resource reservation hook
     */
    public void releaseHook(int pDscp)
    {
	System.out.println("### Called QaSStaticMapper.releaseHook()");
	
	//Update class status
	myQaSClassStatus.myClassTable[myQaSStreamStatus.myStreamTable[pDscp].myDiffServDSCP].allocated -= myQaSStreamStatus.myStreamTable[pDscp].myRate;
    }
    
    
    /*
     * observeHTML(): built an HTML status page
     */
    public void observeHTML(StringBuffer pString)
    {
	//Add Mapper info
	pString.append("<p><bold>Mapping:</bold> VTHD static<p>");
	
	//Add diffserv class status info
	myQaSClassStatus.observeHTML(pString);
	
	//Add stream info
	myQaSStreamStatus.observeHTML(pString);
    }
} // class QaSStaticMapper


/*****************************************************************************/
/* QaSDynPickParam class:                                                    */
/*                                                                           */
/*****************************************************************************/
class QaSDynPickParam extends QaSPickParam
{
    /*************************************************************************/
    /* Public                                                                */
    /*************************************************************************/
    
    float monRes;
    
    
    /*
     * Constructor
     */
    public QaSDynPickParam(int aDscp, float aMonRes)
    {
	super(aDscp);

	this.monRes = aMonRes;
    }
}


/*****************************************************************************/
/* QaSDynamicMapper1 class:                                                  */
/*                         Simple dynamic 1 class mapping algorithm.         */
/*****************************************************************************/
class QaSDynamicMapper1 extends QaSMapper
{
    static final int QAS_EF = 46; //###FC - TBD: this should be read from conf file
    
    
    /*************************************************************************/
    /* Private                                                               */
    /*************************************************************************/
    
    /*
     * mapQualitativeValues(): Translate qualitative guarantee requests into quantitative ones
     */
    void mapQualitativeValues(QaSXMLDoc pDocument)
    {
	//Replace qualitative values with quantitative ones
	pDocument.replaceAttributeByValue("guarantees/delay/*", "high", myXMLConfDoc.getAttribute("qualitative_mapping/delay/high"));
	pDocument.replaceAttributeByValue("guarantees/delay/*", "medium", myXMLConfDoc.getAttribute("qualitative_mapping/delay/medium"));
	pDocument.replaceAttributeByValue("guarantees/delay/*", "low", myXMLConfDoc.getAttribute("qualitative_mapping/delay/low"));
	
	pDocument.replaceAttributeByValue("guarantees/rate/*", "high", myXMLConfDoc.getAttribute("qualitative_mapping/rate/high"));
	pDocument.replaceAttributeByValue("guarantees/rate/*", "medium", myXMLConfDoc.getAttribute("qualitative_mapping/rate/medium"));
	pDocument.replaceAttributeByValue("guarantees/rate/*", "low", myXMLConfDoc.getAttribute("qualitative_mapping/rate/low"));
	
	pDocument.replaceAttributeByValue("guarantees/loss/*", "high", myXMLConfDoc.getAttribute("qualitative_mapping/loss/high"));
	pDocument.replaceAttributeByValue("guarantees/loss/*", "medium", myXMLConfDoc.getAttribute("qualitative_mapping/loss/medium"));
	pDocument.replaceAttributeByValue("guarantees/loss/*", "low", myXMLConfDoc.getAttribute("qualitative_mapping/loss/low"));
    }
    
    
    /*
     * pickDiffServClass(): pick the best DiffServ class corresponding to the request
     */
    QaSDynPickParam pickDiffServClass(QaSXMLDoc pDocument)
    {
	int diffServClass = 0;
	int delayReq;
	int rateReq;
	String aString;
	String ipDest;
	float monRes;
	QaSDynPickParam aParam;
	
	
	//Read requested maximum delay
	aString = pDocument.getAttribute("guarantees/delay/value");
	
	if ( aString.length() != 0 )
	    delayReq = Integer.parseInt(aString); //###FC - TBD: error checking (catch exception)
	else
	    {
		//No delay requested: BE
		aParam = new QaSDynPickParam(0, 0);
		return(aParam);
	    }
	
	//Get monitoring performance of BE
	ipDest = pDocument.getAttribute("flowid/destination");
	monRes = myMonitoring.getClassPerformance(0, ipDest);
	
	if ( monRes > delayReq * 2 )
	    //If BE isn't fast enough, try EF
	    diffServClass = QAS_EF;
	
	System.out.println("IP destination: " + ipDest + "; monitoring: " + monRes / 2 + "; requested: " + delayReq);
	
	aParam = new QaSDynPickParam(diffServClass, monRes);
	
	//Done!
	return(aParam);
    }
    
    
    /*************************************************************************/
    /* Public                                                                */
    /*************************************************************************/
    
    QaSMonitoringIf myMonitoring; //Monitoring interface
    
    
    /*
     * Constructors
     */
    public QaSDynamicMapper1(QaSXMLDoc pXMLConfDoc)
    {
	super(pXMLConfDoc);
	myMonitoring = new QaSMonitoringIf(myXMLConfDoc.getAttribute("monitoring_server/ip"), Integer.parseInt(myXMLConfDoc.getAttribute("monitoring_server/port")));
    }
    
    
    /*
     * initHook(): initialisation hook
     */
    public void initHook()
    {	
	int ret;
	int totalBW;
	int i;
	String aCommand;
	String aString;
	String aClassName;
	
	
        //System.out.println("### Called QaSStaticMapper.initHook()");
	
	for ( i = 0; i < QaSClassStatus.QAS_MAX_NUM_CLASSES; i++ )
	    {
		//Just clean up any leftover module configuration 
		QaSKernelParam param = new QaSKernelParam();
		param.local = i;
		param.diffserv = i;
		param.bucketrate = Integer.parseInt(myXMLConfDoc.getAttribute("configured/link_bandwidth/value")); //In Mbps
		ret = QaSKernelIf.setNetfilter(param);
		aCommand = myXMLConfDoc.getAttribute("token_bucket/script");
		ret = QaSKernelIf.setTc(param, aCommand); //###FC - TBD: get burst and backlog

		
		//Reserve DSCP to allow QoSINUS override
		aClassName = myXMLConfDoc.getFilteredAttribute("configured/class", "dscp", Integer.toString(i), "name");
		
		if ( aClassName.compareTo("") != 0 )
		    {
			myQaSStreamStatus.forceStream(i, false, null, i, aClassName, 0, 0, 0);
		    }
	    }

    }
    
    
    /*
     * acceptHook(): acceptability hook
     */
    public QaSPickParam acceptHook(QaSXMLDoc pDocument)
    {
	int diffServDSCP;
	String ipDest;
	QaSDynPickParam pickRet;
	
	
	//System.out.println("### Called QaSDynamicMapper1.acceptHook()");
	
	//Start class monitoring
	ipDest = pDocument.getAttribute("flowid/destination");
	myMonitoring.startClassMonitoring(0, ipDest);
	
	//Map qualitative values into quantitative ones
	mapQualitativeValues(pDocument);
	
	//Policy checking could come here.
	
	//Pick the best DiffServ class
	pickRet = pickDiffServClass(pDocument);
	
	System.out.println("Picked DiffServ=" + pickRet.dscp);
	
	return(pickRet);
    }
    
    
    /*
     * reserveHook(): resource reservation hook
     */
    public void reserveHook(QaSXMLDoc pDocument, int pLocal, int pChosenClass)
    {
	String aString;
	int delayReq;
	
	
	//System.out.println("### Called QaSDynamicMapper1.reserveHook()");
	
	//Get the rate requested
	aString = pDocument.getAttribute("guarantees/delay/value");
	
	if ( aString.length() != 0 )
	    delayReq = Integer.parseInt(aString); //###FC - TBD: error checking (catch exception)
	else
	    //Use default rate
	    delayReq = Integer.parseInt(myXMLConfDoc.getAttribute("defaults/delay/value")); //###FC - TBD: error handling
	
	//Update stream status
	myQaSStreamStatus.myStreamTable[pLocal].myDelay = delayReq;
    }
    
    
    /*
     * releaseHook(): resource liberation hook
     */
    public void releaseHook(int pDscp)
    {
	String ipDest;
	QaSXMLDoc aDocument;
	
	
	//System.out.println("### Called QaSDynamicMapper1.releaseHook()");
	
	//Retreive the destination in the original SLS
	aDocument = myQaSStreamStatus.myStreamTable[pDscp].mySLS;
	
	if ( aDocument != null )
	    {
		ipDest = aDocument.getAttribute("flowid/destination");
		
		//Stop class monitoring
		myMonitoring.stopClassMonitoring(0, ipDest);
	    }
    }
    
    
    /*
     * heartBeatHook(): pulse hook
     */
    public void heartBeatHook()
    {
	int i;
	int ret;
	QaSKernelParam param = new QaSKernelParam();
	QaSDynPickParam pickRet;
	String aClassName;
	
	//System.out.println("### Called dynamic heartbeat hook.");
	
	//Traverse the stream table
	for ( i = 0; i < myQaSStreamStatus.QAS_MAX_NUM_STREAMS; i++ )
	    {
		if ( myQaSStreamStatus.myStreamTable[i].available == false && myQaSStreamStatus.myStreamTable[i].mySLS != null )
		    {
			pickRet = pickDiffServClass(myQaSStreamStatus.myStreamTable[i].mySLS);
			
			myQaSStreamStatus.myStreamTable[i].monRes = pickRet.monRes;
			
			if ( myQaSStreamStatus.myStreamTable[i].myDiffServDSCP != pickRet.dscp )
			    {
				//The stream is assigned a new class
				myQaSStreamStatus.myStreamTable[i].myDiffServDSCP = pickRet.dscp;
				aClassName = myXMLConfDoc.getFilteredAttribute("configured/class", "dscp", Integer.toString(pickRet.dscp), "name");
				myQaSStreamStatus.myStreamTable[i].myDiffServName = aClassName;
				
				//Configure kernel module
				param.local = i;
				param.diffserv = pickRet.dscp;
				//param.bucketrate = myQaSStreamStatus.myStreamTable[i].myRate;
				ret = QaSKernelIf.setNetfilter(param);

				System.out.println("### Set DSCP " + param.diffserv + " for stream " + param.local);
			    }
		    }    
	    }
    }
    
    
    /*
     * observeHTML(): built an HTML status page
     */
    public void observeHTML(StringBuffer pString)
    {
	//Add Mapper info
	pString.append("<table border=\"0\"><tbody><tr><td>Mapping: VTHD dynamic</td></tr><tr><td>  </td></tr></tbody></table>");
	
	//Add diffserv class status info
	//myQaSClassStatus.observeHTML(pString);
	
	//Add stream info
	myQaSStreamStatus.observeHTML(pString);
    }
} // class QaSDynamicMapper1




