/*****************************************************************************/
/*                                                                           */
/*  QaSMonitoringSrv                                                         */
/*                                                                           */
/*                                                                           */
/*     Description:                                                          */
/*             The DiffServ classes monitoring server.                       */
/*                                                                           */
/*     History:                                                              */
/*             14/04/2003 - Fabien Chanussot - Creation                      */
/*                                                                           */
/*****************************************************************************/

//###FC - TBD package QaS;

/*****************************************************************************/
/* Use                                                                       */
/*****************************************************************************/

import java.util.*;
import java.io.*;
import java.net.*;

import org.xml.sax.InputSource;

/*****************************************************************************/
/* QaSNodeListHead class:                                                    */
/*                        Linked list of destination being monitored         */
/*****************************************************************************/
class QaSMonitoringListHeader
{
    /*************************************************************************/
    /* Public                                                                */
    /*************************************************************************/
    
    QaSMonitoringListNode next;
    
    
    /*
     * Constructor:
     */
    QaSMonitoringListHeader()
    {
	next = null;
    }
    
    
    /*
     * get(): scan the list looking for a IP being monitored
     */
    QaSMonitoringListNode get(InetAddress ipSearched)
    {
	if ( next == null )
	    {
		//Empty list: destination not found
		return(null);
	    }
	else
	    {
		//Iteration
		return(next.get(ipSearched));
	    }
    }
    
    
    /*
     * add(): add a new destination monitoring to the list
     */
    QaSMonitoringListNode add(InetAddress newIp)
    {
	if ( next == null )
	    {
		//End of the list: append new element
		next = new QaSMonitoringListNode(newIp);
		return(next);
	    }
	else
	    {
		//Iteration
		return(next.add(newIp));
	    }
    }
    
    
    /*
     * remove(): remove a destination monitoring entry from the list
     */
    QaSMonitoringListNode remove(InetAddress ipSearched)
    {
	if ( next != null )
	    {
		//Iteration
		return(next.remove(this, ipSearched));
	    }
	else
	    {
		return(null);
	    }
    }
}


/*****************************************************************************/
/* QaSNodeListNode class:                                                    */
/*                        Monitoring linked list node                        */
/*****************************************************************************/
class QaSMonitoringListNode extends QaSMonitoringListHeader
{
    /*************************************************************************/
    /* Public                                                                */
    /*************************************************************************/
    
    InetAddress ipAddr;
    QaSMonitoringThread rttThread;
    int reference;
    
    
    /*
     * Constructor:
     */
    QaSMonitoringListNode(InetAddress newIp)
    {
	ipAddr = newIp;
	rttThread = null;
	reference = 1;
    }
    
    
    /*
     * get(): scan the list looking for a IP being monitored
     */
    QaSMonitoringListNode get(InetAddress ipSearched)
    {
	if ( ipAddr.equals(ipSearched) )
	    {
		//The destination is already monitored
		return(this);
	    }
	else
	    {
		if ( next == null )
		    {
			//End of the list: destination not found
			return(null);
		    }
		else
		    {
			//Iteration
			return(next.get(ipSearched));
		    }
	    }
    }
    
    
    /*
     * add(): add a new destination monitoring to the list
     */
    QaSMonitoringListNode add(InetAddress newIp)
    {
	if ( ipAddr.equals(newIp) )
	    {
		//The destination is already monitored
		reference = reference + 1;
		return(this);
	    }
	else
	    {
		if ( next == null )
		    {
			//End of the list: append new element
			next = new QaSMonitoringListNode(newIp);
			return(next);
		    }
		else
		    {
			//Iteration
			return(next.add(newIp));
		    }
	    }
    }
    
    
    /*
     * remove(): remove a destination monitoring entry from the list
     */
    QaSMonitoringListNode remove(QaSMonitoringListHeader previous, InetAddress ipSearched)
    {
	if ( ipAddr.equals(ipSearched) )
	    {
		//Decrement reference
		reference = reference - 1;
		
		if ( reference == 0 )
		    {
			//Last reference: remove the node
			previous.next = next;
		    }
		
		return(this);
	    }
	else
	    {
		if ( next != null )
		    {
			//Iteration
			return(next.remove(this, ipSearched));
		    }
		else
		    {
			return(null);
		    }
	    }
    }

    
    /*
     * start(): arrancar la tarea de monitoring
     */
    void start(int aClass, String aCommand, int aDelay)
    {
    
	rttThread = new QaSMonitoringThread(aClass, aCommand, aDelay);
	rttThread.start();
    }
    
    
    /*
     * stop(): parar la tarea de monitoring
     */
    void stop()
    {
	if ( rttThread != null )
	    {
		rttThread.killMe = 1;
		rttThread = null;
	    }
    }
}


/*****************************************************************************/
/* QaSMonProcStdStreamer class:                                              */
/*                              Parse the monitoring tool's stderr/stdout    */
/*****************************************************************************/
class QaSMonProcStdStreamer extends Thread
{
    /*************************************************************************/
    /* Public                                                                */
    /*************************************************************************/
    
    InputStream myStream;
    String myType;
    QaSMonitoringThread myMonThread;
    
    
    /*
     * Constructor:
     */
    QaSMonProcStdStreamer(InputStream anInputStream, String aType, QaSMonitoringThread aMonThread)
    {
	this.myStream = anInputStream;
	this.myType = aType;
	this.myMonThread = aMonThread;
    }
    
    
    /*
     * run(): guess what
     */
    public void run()
    {
	if ( myType.compareTo("stdout") == 0 ) 
	    {
		try
		    {
			InputStreamReader anInputReader = new InputStreamReader(myStream);
			BufferedReader aBuffReader = new BufferedReader(anInputReader);
			String aLine = null;
			boolean foundres = false;
			
			while ( (aLine = aBuffReader.readLine()) != null )
			    {
				if ( aLine.startsWith("rtt") )
				    {
					StringTokenizer aTokenizer = new StringTokenizer(aLine, "=");
					
					aTokenizer.nextToken();
					
					aTokenizer.nextToken("=/ ");
					Float aFloat = new Float(aTokenizer.nextToken());

					synchronized ( myMonThread )
					    {
						myMonThread.monResult = aFloat.floatValue();
					    }
					foundres = true;
				    }
			    }
			
			if ( foundres == false )
			    {
				synchronized ( myMonThread )
				    {
					myMonThread.monResult = 20000; //###FC - TBD: make that clean
				    }
				foundres = true;
			    }
		    }
		catch (IOException ioe)
		    {
			ioe.printStackTrace();
		    }
	    }
	else if ( myType.compareTo("stderr") == 0 )
	    {
		try
		    {
			InputStreamReader anInputReader = new InputStreamReader(myStream);
			BufferedReader aBuffReader = new BufferedReader(anInputReader);
			String aLine = null;
			
			while ( (aLine = aBuffReader.readLine()) != null )
			    {
				//An error occured
				myMonThread.error = true;
			    }
		    }
		catch (IOException ioe)
		    {
			ioe.printStackTrace();
		    }
	    }
    }
}


/*****************************************************************************/
/* QaSMonitoringThread class:                                                */
/*                           The working thread dealing with a class/metric. */
/*****************************************************************************/
class QaSMonitoringThread extends Thread
{
    /*************************************************************************/
    /* Public                                                                */
    /*************************************************************************/
    
    int monClass; //DiffServ class being monitorized
    String monCommand; //Command executed to monitor the class
    float monResult; //Instantaneous monitoring result
    int monLoopDelay; //Delay between loop iterations
    int killMe = 0; //Used to stop the thread
    boolean error = false; //Flag that's raised when an error happens
    
    
    /*
     * Constructor:
     */
    public QaSMonitoringThread(int aClass, String aCommand, int aDelay)
    {
	monClass = aClass;
	monResult = 0; //###FC - TBD: get default from conf file
	monCommand = aCommand;
	monLoopDelay = aDelay;
    }
    
    
    /*
     * run(): pues eso...
     */
    public void run()
    {
	Runtime aRunTime = Runtime.getRuntime();
	
	while ( killMe == 0 )
	    {
		try
		    {
			Process aProc = aRunTime.exec(monCommand);
			
			QaSMonProcStdStreamer aStrErr = new QaSMonProcStdStreamer(aProc.getErrorStream(), "stderr", this);
			QaSMonProcStdStreamer aStrOut = new QaSMonProcStdStreamer(aProc.getInputStream(), "stdout", this);
			
			aStrErr.start();
			aStrOut.start();
			
			System.out.print("<");			
			aStrErr.join();
			aStrOut.join();
			System.out.println(">");
			
			int iret = aProc.waitFor();
			
			if ( iret != 0 )
			    System.err.println("QaSMonitoring waitFor returned: " + iret);
			
			try
			    {
				sleep(monLoopDelay);
				if ( error == false ) 
				    System.out.println(monCommand + " : " + monResult );
				else
				    {
					System.out.println("Error detected");
					//###FC - TBD: set result to err value.
					error = false;
				    }
			    }
			catch (Exception e)
			    {
				System.err.println("QaSMonitoring err: " + e.getMessage());
			    }
		    }
		catch (Throwable t)
		    {
			t.printStackTrace();
		    }
	    }
    }
}


/*****************************************************************************/
/* QaSMonConnection class:                                                   */
/*                         Class that handle a connection with a client      */
/*****************************************************************************/
class QaSMonConnection extends Thread
{
    /*************************************************************************/
    /* Protected                                                             */
    /*************************************************************************/
    
    protected Socket client;
    protected DataInputStream in;
    protected DataOutputStream out;
    protected QaSMonitoringSrv monitoringSrv;
    
    
    /*************************************************************************/
    /* Public                                                                */
    /*************************************************************************/
    
    /*
     * Constructors
     */
    public QaSMonConnection(Socket clientSocket, QaSMonitoringSrv aServer)
    {
	this.client = clientSocket;
	this.monitoringSrv = aServer;
	
	try
	    {
		in = new DataInputStream(client.getInputStream());
		out = new DataOutputStream(client.getOutputStream());
	    }
	catch (IOException e)
	    {
		try {this.client.close();} catch (IOException e2){}
		System.err.println("Error: exception while getting socket streams: " + e);
	    }
	
	/* A new connection arrived; handle it */
	this.start();
    }
    
    
    /*
     * run(): read the request, serve it and send response back to the client
     */
    public void run()
    {
	int op, diffServ, tool;
	String dest;
	InetAddress destInet;
	float monRes = -1;
	
	
	try
	    {
		//Read operation, class and tool
		op = in.readInt();
		diffServ = in.readInt();
		//###FC - TBD tool = in.readInt();
		
		//Read ip address
		dest = in.readLine();
		destInet = InetAddress.getByName(dest);
		
		//Run the corresponding service
		switch ( op )
		    {
		    case 1:
			//Start monitoring
			System.out.println("Received start req.");
			
			monitoringSrv.startClassMonitoring(diffServ, 0, destInet);
			break;
		    case 2:
			//Stop monitoring
			System.out.println("Received stop req.");
			
			monitoringSrv.stopClassMonitoring(diffServ, destInet);
			break;
		    case 3:
			//Get monitoring result
			System.out.println("Received get req.");
			
			monRes = monitoringSrv.getClassPerformance(diffServ, destInet);
			break;
		    }
		
		//Send response back to the client
		if ( op == 3 )
		    {
			out.writeFloat(monRes);
		    }
		else
		    {
			out.writeInt(0); //sync client and server
		    }
	    }
	catch (IOException e)
	    {
		System.err.println("Error: exception while servicing request: " + e);
	    }
    }
}


/*****************************************************************************/
/* QaSMonitoringSrv class:                                                   */
/*                        The DiffServ classes monitoring class.             */
/*****************************************************************************/
class QaSMonitoringSrv extends Thread
{
    /*************************************************************************/
    /* Private                                                               */
    /*************************************************************************/
    
    private QaSXMLIf  myQaSXMLIf; //Interface with XML parser
    QaSXMLDoc myXMLConfDoc; //Configuration file

    
    
    /*************************************************************************/
    /* Protected                                                             */
    /*************************************************************************/
    
    //Maximum number of classes. The DSCP is coded on a 6-bits value
    protected static final int QAS_MAX_NUM_CLASSES = 256; //###FC - Fixe that:-(((
    
    //Port the server is listening
    protected int port;
    
    //Server socket
    protected ServerSocket listenSocket;
    
    
    /*************************************************************************/
    /* Public                                                                */
    /*************************************************************************/
    
    public QaSMonitoringListHeader myMonitoringTable[];
    
    
    /*
     * Constructor:
     */
    public QaSMonitoringSrv(String args[]) throws IOException
    {
	int i;
	FileInputStream aStream;
	InputSource aSource;
	
	
	//Initialize monitoring linked lists
	myMonitoringTable = new QaSMonitoringListHeader[QAS_MAX_NUM_CLASSES];
	
	for ( i = 0; i < QAS_MAX_NUM_CLASSES; i++ )
	    {
		myMonitoringTable[i] = new QaSMonitoringListHeader();
	    }
	
	//Conf file
	myQaSXMLIf = new QaSXMLIf();
	
	//Parse the configuration XML document
	aStream = new FileInputStream(args[0]);
	aSource = new InputSource(aStream);
	myXMLConfDoc = myQaSXMLIf.parse(aSource);
	
	//Create socket server
	this.port = Integer.parseInt(myXMLConfDoc.getAttribute("server/port"));
	
	try
	    {
		this.listenSocket = new ServerSocket(this.port);
	    }
	catch (IOException e)
	    {
		fail(e, "Error: exception creating server socket");
	    }
	
	/* Start the server */
	this.start();
    }
    
    
    /*
     * fail(): deal with errors
     */
    public void fail(Exception e, String message)
    {
	System.err.println(message + ": " + e);
	System.exit(1);
    }
    
    
    /*
     * run(): loop listening for and accepting new connections
     */
    public void run()
    {
	try
	    {
		while (true)
		    {
			/* Listen for incoming connections */
			Socket clientSocket = listenSocket.accept();
			
			/* Handle communication */
			QaSMonConnection aMonConnection = new QaSMonConnection(clientSocket, this);
		    }
	    }
	catch (IOException e)
	    {
		fail(e, "Error: exception while listening");
	    }
    }
    
    
    /*
     * startClassMonitoring(): launch a thread to monitor performance of a class
     */
    public synchronized void startClassMonitoring(int aClass, int aType, InetAddress destIp)
    {
	QaSMonitoringListNode aNode;
	String aCommand;
	int startDelay;
	int loopDelay;
	
	
	aNode = myMonitoringTable[aClass].add(destIp);
	
	if ( aNode.reference == 1 )
	    {
		//Read monitoring command from conf file
		aCommand = myXMLConfDoc.getFilteredAttribute("tool", "type", Integer.toString(aType), "script");
		loopDelay = Integer.parseInt(myXMLConfDoc.getFilteredAttribute("tool", "type", Integer.toString(aType), "loop_delay"));
		
		aNode.start(aClass, aCommand + destIp.getHostAddress(), loopDelay);
		
		try 
		    {
			//Sleep to give the thread time to get its first result
			startDelay = Integer.parseInt(myXMLConfDoc.getFilteredAttribute("tool", "type", Integer.toString(aType), "start_delay"));
			sleep(startDelay);
		    }
		catch (Exception e)
		    {
			System.err.println("QaSMonitoring err: " + e.getMessage());
		    }
	    }
    }
    
    
    /*
     * stopClassMonitoring(): stop a monitoring thread
     */
    public synchronized void stopClassMonitoring(int aClass, InetAddress destIp)
    {
	QaSMonitoringListNode aNode;
	
	aNode = myMonitoringTable[aClass].remove(destIp);
	
	if ( aNode != null )
	    {
		if ( aNode.reference <= 0 )
		    aNode.stop();
	    }
    }
    
    
    /*
     * getClassPerformance(): retreive a performance metric
     */
    public synchronized float getClassPerformance(int aClass, InetAddress aDestIp)
    {
	QaSMonitoringListNode aNode;
	
	aNode = myMonitoringTable[aClass].get(aDestIp);
	
	if ( aNode == null )
	    {
		return(-1);
	    }
	else
	    {
		float res;
		
		synchronized ( aNode.rttThread )
		    {
			res = aNode.rttThread.monResult;
		    }
		return(res);
	    }
    }
    
    
    /*
     * main(): pues eso, main
     */
    public static void main(String args[])
    {
	QaSMonitoringSrv theMonitoringServer;
	
	//Parse arguments
	if ( args.length != 1 )
	    {
       		System.out.println ("Invalid arguments: QaSMonitoringSrv conf_file.xml");
      		System.exit (1);
       	    }
	
	//Create the monitoring threads server
	try 
	    {
		theMonitoringServer = new QaSMonitoringSrv(args);
	    }
	catch (IOException e)
	    {
		System.err.println("Error: exception while geting socket streams: " + e);
	    }
 
    }
}


