/*****************************************************************************/
/*                                                                           */
/*  QaSStreamStatus                                                          */
/*                                                                           */
/*                                                                           */
/*     Description:                                                          */
/*             The table containing information about streams.               */
/*                                                                           */
/*     History:                                                              */
/*             30/01/2003 - Fabien Chanussot - Creation                      */
/*                                                                           */
/*****************************************************************************/

//###FC - TBD package QaS;

/*****************************************************************************/
/* Use                                                                       */
/*****************************************************************************/

import java.util.*;
import java.io.*;


/*****************************************************************************/
/* QaSStreamEntry class:                                                     */
/*                         Status information about a stream.                */
/*****************************************************************************/
class QaSStreamEntry
{
    /*************************************************************************/
    /* Private                                                               */
    /*************************************************************************/
    
    boolean available;          //Is this entry available?
    QaSXMLDoc mySLS;            //SLS requested for that stream
    int myDiffServDSCP;         //DiffServ class to be used for that stream
    String myDiffServName;      //and its name
    int mySoftStateTimeout;     //Number of timer pulses left before the soft state releases the stream's ressources
    int mySoftStateGracePeriod; //Initial value of mySoftStateTimeout
    int myRate;                 //Rate reserved for that stream ###FC - TBD: this is specific to StaticMapper
    int myDelay;                //Delay max requested for that stream ###FC - TBD: this is specific to DynamicMapper1
    float monRes;              //Delay monitored ###FC - TBD: this is specific to DynamicMapper1
    
    
    /*************************************************************************/
    /* Public                                                                */
    /*************************************************************************/
    
    /*
     * Constructor:
     */
    public QaSStreamEntry()
    {
	available = true;
	myDiffServDSCP = 0;
	mySLS = null;
	mySoftStateTimeout = 0;
	myRate = 0;
	myDelay = 0;
	monRes = 0;
    }
    
    /*
     * set: store information about a new stream in that entry
     */
    public void set(QaSXMLDoc pSLS, int pDiffServDSCP, int pTimeout, String pName)
    {
	myDiffServDSCP = pDiffServDSCP;
	myDiffServName = pName;
	mySLS = pSLS;
	mySoftStateGracePeriod = pTimeout;
	mySoftStateTimeout = pTimeout;
    }
    
    
    /*
     * observeHTML(): built an HTML status page
     */
    public void observeHTML(StringBuffer pString, int i)
    {
	pString.append("<tr><td><h3>");
	pString.append(i);
	pString.append("</h3></td><td><h3>");
	pString.append(this.myDiffServName);
	pString.append("</h3></td><td><h3>");
	pString.append(this.myDelay + " ms");
	pString.append("</h3></td><td><h3>");
	pString.append(this.monRes + " ms");
	pString.append("</h3></td></tr>");
    }
}


/*****************************************************************************/
/* QaSStreamStatus class:                                                    */
/*                         Status information table.                         */
/*****************************************************************************/
class QaSStreamStatus
{
    /*************************************************************************/
    /* Private                                                               */
    /*************************************************************************/
    
    /*************************************************************************/
    /* Public                                                                */
    /*************************************************************************/
    
    //Maximum number of streams that we can deal with in parallel. The DSCP is coded on a byte.
    static final int QAS_MAX_NUM_STREAMS = 256;
    
    //Streams table
    public QaSStreamEntry myStreamTable[];

    
    /*
     * Constructor:
     */
    public QaSStreamStatus()
    {
	int i;
	
	myStreamTable = new QaSStreamEntry[QAS_MAX_NUM_STREAMS];
	
	for ( i = 0; i < QAS_MAX_NUM_STREAMS; i++ )
	    myStreamTable[i] = new QaSStreamEntry();
	
	//###FC - TBD: mark the reserved DSCP values (passthru mechanism).
    }
    
    
    /*
     * addStream(): add a new stream in the table
     */
    public int addStream(QaSXMLDoc pSLS, int pDiffServDSCP, String pDiffServName, int pTimeout)
    {
	int i;
	boolean found = false;
	
	//Find the first available spot and use it
	for ( i = 0; i < QAS_MAX_NUM_STREAMS && found == false; i++ )
	    { //Note: entry 0 is reserved (best effort). We keep the entry in the table just in case.
		if ( myStreamTable[i].available )
		    { //We've found a spot
			found = true;
			myStreamTable[i].available = false;
			myStreamTable[i].mySLS = pSLS;
			myStreamTable[i].myDiffServDSCP = pDiffServDSCP;
			myStreamTable[i].myDiffServName = pDiffServName;
			myStreamTable[i].mySoftStateGracePeriod = pTimeout;
			myStreamTable[i].mySoftStateTimeout = pTimeout;
		    }
	    }
	
	if ( found )
	    return(i - 1);
	else
	    return(-1);
    }
    
    
    /*
     * forceStream(): force the value of an entry in the table
     */
    public void forceStream(int pIndex, boolean pAvailable, QaSXMLDoc pSLS, int pDiffServDSCP, String pDiffServName, int pTimeout, int pRate, int pDelay)
    {
	myStreamTable[pIndex].available = pAvailable;
	myStreamTable[pIndex].mySLS = pSLS;
	myStreamTable[pIndex].myDiffServDSCP = pDiffServDSCP;
	myStreamTable[pIndex].myDiffServName = pDiffServName;
	myStreamTable[pIndex].mySoftStateGracePeriod = pTimeout;
	myStreamTable[pIndex].mySoftStateTimeout = pTimeout;
	myStreamTable[pIndex].myRate = pRate;
	myStreamTable[pIndex].myDelay = pDelay;
    }
    
    
    /*
     * removeStream(): free an entry in the list
     */
    public void removeStream(int pLocalDSCP)
    {
	//###FC - TBD: sanity check
	
	//Mark the entry as available
	myStreamTable[pLocalDSCP].available = true; //###FC - TBD: check type
	myStreamTable[pLocalDSCP].mySLS = null;
	myStreamTable[pLocalDSCP].myDiffServDSCP = 0;
	myStreamTable[pLocalDSCP].mySoftStateGracePeriod = 0;
	myStreamTable[pLocalDSCP].mySoftStateTimeout = 0;
	myStreamTable[pLocalDSCP].myRate = 0;
	myStreamTable[pLocalDSCP].myDelay = 0;
	myStreamTable[pLocalDSCP].monRes = 0;
    }
    
    
    /*
     * pulse: decrement the soft state counter of an entry
     */
    public boolean pulse(int pLocalDSCP)
    {
	boolean timedout = false;
	
	if ( !myStreamTable[pLocalDSCP].available )
	    if ( myStreamTable[pLocalDSCP].mySoftStateTimeout > 0 )
		{
		    myStreamTable[pLocalDSCP].mySoftStateTimeout--;
		    
		    if ( myStreamTable[pLocalDSCP].mySoftStateTimeout == 0 )
			timedout = true;
		}
	
	return(timedout);
    }
    
    
    /*
     * refresh: reinit the soft state counter
     */
    public void refresh(int pLocalDSCP)
    {
	myStreamTable[pLocalDSCP].mySoftStateTimeout = myStreamTable[pLocalDSCP].mySoftStateGracePeriod;
    }
    
    
    /*
     * dump:
     */
    public void dump()
    {
	int i;
	
	//System.out.println("Dumping Stream table:");
	
	for ( i = 0; i < QAS_MAX_NUM_STREAMS; i++ )
	    {
		if ( ( myStreamTable[i].mySLS != null ) && ( myStreamTable[i].available == false ) )
		    {
			//System.out.println("LocalId: " + i + ", DSCP: " + myStreamTable[i].myDiffServDSCP + ", rate: " + myStreamTable[i].myRate + ", Timeout in: " + myStreamTable[i].mySoftStateTimeout);
			System.out.println("LocalId: " + i + ", DSCP: " + myStreamTable[i].myDiffServDSCP + ", requested delay: " + myStreamTable[i].myDelay + ", Timeout in: " + myStreamTable[i].mySoftStateTimeout);
		    }
	    }
    }
    
    
    /*
     * observeHTML(): built an HTML status page
     */
    public void observeHTML(StringBuffer pString)
    {
	int i;
	boolean empty = true;
	
	
	//Add table
	pString.append("<table cellpadding=\"2\" cellspacing=\"2\" border=\"0\" style=\"width: 100%; text-align: left;\"><tbody>");
	pString.append("<tr align=\"center\" style=\"font-family: helvetica,arial,sans-serif;\"><td style=\"vertical-align: top;\" colspan=\"4\"><h2 style=\"background-color: rgb(0, 0, 255);\"><span style=\"color: rgb(255, 255, 255);\"></span><span style=\"color: rgb(255, 255, 255);\">Streams status</span><br></h2></td></tr>");
	pString.append("<tr><td style=\"vertical-align: top; color: rgb(0, 0, 255);\"><h3 style=\"background-color: rgb(221, 221, 221); font-family: helvetica,arial,sans-serif;\">Stream's ID<br></h3></td><td style=\"vertical-align: top; color: rgb(0, 0, 255);\"><h3 style=\"background-color: rgb(221, 221, 221); font-family: helvetica,arial,sans-serif;\">DiffServ class used<br></h3></td><td style=\"vertical-align: top; color: rgb(0, 0, 255);\"><h3 style=\"background-color: rgb(221, 221, 221); font-family: helvetica,arial,sans-serif;\">Requested delay<br></h3></td><td style=\"vertical-align: top; color: rgb(0, 0, 255);\"><h3 style=\"background-color: rgb(221, 221, 221); font-family: helvetica,arial,sans-serif;\">Observed RTT<br></h3></td></tr>");
	
	for ( i = 0; i < QAS_MAX_NUM_STREAMS; i++ )
	    {
		if ( ( myStreamTable[i].mySLS != null ) && ( myStreamTable[i].available == false ) )
		    {
			myStreamTable[i].observeHTML(pString, i);
			empty = false;
		    }		
	    }
	
	pString.append("</tbody></table>");
	
	if ( empty == true )
	    pString.append("<table border=\"0\"><tbody><tr><td>No stream...</td></tr></tbody></table>");
    }
}
