/*****************************************************************************/
/*                                                                           */
/*  qasfilter.c                                                              */
/*                                                                           */
/*                                                                           */
/*     Module:                                                               */
/*             qasmodule                                                     */
/*                                                                           */
/*     Description:                                                          */
/*             Packets filtering and marking, according to the QoSINUS       */
/*             active service programmation.                                 */
/*                                                                           */
/*     History:                                                              */
/*             11/02/2003 - Fabien Chanussot - Creation                      */
/*                                                                           */
/*****************************************************************************/

#define MODULE 
#define __KERNEL__


/*****************************************************************************/
/* Use                                                                       */
/*****************************************************************************/

#include <linux/config.h>
#include <linux/types.h>
#include <linux/ip.h>
#include <linux/timer.h>
#include <linux/module.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv4/ip_conntrack.h>
#include <net/protocol.h> 
#include <net/checksum.h> 

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <asm/uaccess.h>

#include "qasfilterP.h"
#include "qasmainP.h"
#include "qastbP.h"


/*****************************************************************************/
/* Macros                                                                    */
/*****************************************************************************/

/*****************************************************************************/
/* Types                                                                     */
/*****************************************************************************/

/*****************************************************************************/
/* Global                                                                    */
/*****************************************************************************/

/* Netfilter hook */
static unsigned int qasFilterFilter(unsigned int hooknum,
				    struct sk_buff **pskb,
				    const struct net_device *in,
				    const struct net_device *out,
				    int (*okfn)(struct sk_buff *));

/* Netfilter hooks table */
struct nf_hook_ops qaspostroute_ops = 
{
  { NULL, NULL },
  qasFilterFilter,	/* associate function */
  PF_INET, 		/* IP protocol */
  NF_IP_POST_ROUTING, 	/* hook */
  NF_IP_PRI_FILTER 	/* priority level */
};


/*****************************************************************************/
/* Private functions                                                         */
/*****************************************************************************/

/*****************************************************************************/
/* Function:                                                                 */
/* Role    :                                                                 */
/*****************************************************************************/
static unsigned int qasFilterFilter(unsigned int hooknum,
				    struct sk_buff **pskb,
				    const struct net_device *in,
				    const struct net_device *out,
				    int (*okfn)(struct sk_buff *))
{
    unsigned int prot;
    u_int16_t diffs[2];
    unsigned char pkTos, oldTos;
    
    
    if (*pskb != NULL) 
	{
	    prot = SKB_PROTOCOL_IP(*pskb);  
	    
	    if ( prot == TCP || prot == UDP || prot == ICMP ) //###FC - Check that
		{
		    oldTos = SKB_HEAD_IP(*pskb)->tos;
		    
		    pkTos = oldTos / 4;
		    
		    /* There is activity */
		    qasdev_softstate[pkTos].activity = 1;
		    
		    /* Mark the packet so that it's branched to the correct token bucket (using streamId) */
		    (*pskb)->nfmark = pkTos;
		    
		    /* If the packet's TOS is set and remarking is needed, overwrite the tos field */
		    //###FC - DEBUG if ( ( pkTos != 0 ) && ( qasdev_softstate[pkTos].dscp != pkTos ) )
		    if ( qasdev_softstate[pkTos].dscp != pkTos )
			{
			    //Update checksum
			    diffs[0] = htons(oldTos) ^ 0xFFFF;
			    
			    //Mark the TOS based on soft state table
			    SKB_HEAD_IP(*pskb)->tos = qasdev_softstate[pkTos].dscp * 4;
			    
			    diffs[1] = htons(SKB_HEAD_IP(*pskb)->tos);
			    SKB_HEAD_IP(*pskb)->check = csum_fold(csum_partial((char *)diffs,
									       sizeof(diffs),
									       SKB_HEAD_IP(*pskb)->check^0xFFFF));
			}
		    
		}
	    
	}
    
    
    return(NF_ACCEPT);
}


/*****************************************************************************/
/* Protected functions                                                       */
/*****************************************************************************/

/*****************************************************************************/
/* Public functions                                                          */
/*****************************************************************************/



