/*****************************************************************************/
/*                                                                           */
/*  qasmain.c                                                                */
/*                                                                           */
/*                                                                           */
/*     Module:                                                               */
/*             qasmodule                                                     */
/*                                                                           */
/*     Description:                                                          */
/*             Module that filters and mark the packets according to the     */
/*             QoSINUS active service programmation and provides the token   */
/*             bucket functionality.                                         */
/*             This file implements the pseudo device used to communicate    */
/*             with the userland Java actif service, and is the module's     */
/*             entry point.                                                  */
/*                                                                           */
/*     History:                                                              */
/*             27/11/2002 - Fabien Chanussot - Creation                      */
/*                          heavily based on Rusty Russel and JPGelas code.  */
/*                                                                           */
/*****************************************************************************/

#define MODULE 
#define __KERNEL__


/*****************************************************************************/
/* Use                                                                       */
/*****************************************************************************/

#include <linux/config.h>
#include <linux/types.h>
#include <linux/ip.h>
#include <linux/timer.h>
#include <linux/module.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv4/ip_conntrack.h>
#include <net/protocol.h> 
#include <net/checksum.h> 

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <asm/uaccess.h>

#include "qasmainP.h"
#include "qasfilterP.h"
#include "qastbP.h"


/*****************************************************************************/
/* Macros                                                                    */
/*****************************************************************************/

/*****************************************************************************/
/* Types                                                                     */
/*****************************************************************************/

/*****************************************************************************/
/* Global                                                                    */
/*****************************************************************************/

/* qasdev softstate table */
qasSoftState_t *qasdev_softstate;

/* Fonctions declaration */
static int qasdev_open(struct inode *inode, struct file *file);
static int qasdev_release(struct inode *inode, struct file *file);
static ssize_t qasdev_read(struct file *file, char *buf, size_t count, loff_t *ppos);
static ssize_t qasdev_write(struct file *file, const char *buf, size_t count, loff_t *ppos);
static loff_t qasdev_llseek(struct file *file, loff_t offset, int whence);

/* qasdev major number */
static unsigned int qasdev_major = 0;

/* Pseudo device hooks */
static struct file_operations qasdev_fops =
{
  THIS_MODULE,     /* struct module owner */
  qasdev_llseek,   /* qasdev_llseek       */
  qasdev_read,     /* qasdev_read         */
  qasdev_write,    /* qasdev_write        */
  NULL,            /* qasdev_readdir      */
  NULL,            /* qasdev_poll         */
  NULL,            /* qasdev_ioctl        */
  NULL,            /* qasdev_mmap         */
  qasdev_open,     /* qasdev_open         */
  NULL,            /* qasdev_flush        */
  qasdev_release,  /* qasdev_release      */
  NULL,            /* qasdev_fsync        */
  NULL,            /* qasdev_fasync       */
  NULL,            /* qasdev_lock         */
  NULL,            /* qasdev_readv        */
  NULL             /* qasdev_writev       */
};


/*****************************************************************************/
/* Private functions                                                         */
/*****************************************************************************/

/*****************************************************************************/
/* Protected functions                                                       */
/*****************************************************************************/

/*****************************************************************************/
/* Public functions                                                          */
/*****************************************************************************/

/*****************************************************************************/
/* Function: qasdev_open                                                     */
/* Role    : Called when the pseudo device is opened.                        */
/*****************************************************************************/
static int qasdev_open(struct inode *inode, struct file *file)
{
  //printk("QoSINUS: qasdev_open called\n");
  
  return(0);
}


/*****************************************************************************/
/* Function: qasdev_release                                                  */
/* Role    : Called when the pseudo device is closed.                        */
/*****************************************************************************/
static int qasdev_release(struct inode *inode, struct file *file)
{
  //printk("QoSINUS: qasdev_release called\n");
  //###TBD - MOD_DEC_USE_COUNT
  
  return(0);
}


/*****************************************************************************/
/* Function: qasdev_read                                                     */
/* Role    : Called when the pseudo device is read.                          */
/*****************************************************************************/
static ssize_t qasdev_read(struct file *file, char *buf, size_t count, loff_t *ppos)
{
    int ret;
    char *ptByte;
    int32_t longint, longint2;
    
    
    //###FC - DEBUG printk("QoSINUS: qasdev_read called. Offset: %d, count: %d", *ppos, count);
    
    /* Check offset */
    if ( *ppos >= QAS_MAX_NUM_STREAMS * sizeof(qasSoftState_t) )
	{
	    printk("QoSINUS! Trying to read at the end of file.\n");
	    return(-ESPIPE);
	}
    
    /* This kludge only works when reading 4-Byte integers byte by byte */
    if ( count != 1 )
	{
	    printk("QoSINUS! Trying to read at the end of file.\n");
	    return(-ESPIPE);
	}
    
    /* Get the 32 bit int being read */
    longint = ((int32_t *)qasdev_softstate)[*ppos / sizeof(int32_t)];
    
    /* Swap it to network byte order */
    longint2 = htonl(longint);
    
    /* Get the byte being read */
    ptByte = &((char *)&longint2)[*ppos % sizeof(int32_t)];
    
    /* Copy byte to userland */
    ret = copy_to_user(buf, ptByte, count);
    
    if ( ret != 0 )
	return(-EFAULT);

    /* Update offset */
    *ppos = *ppos + count;
    
    
    /* Done! */
    return(count);
}


/*****************************************************************************/
/* Function: qasdev_write                                                    */
/* Role    : Called when the pseudo device is written.                       */
/*****************************************************************************/
static ssize_t qasdev_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
    int ret;
    char *ptr;
    int32_t longint, *ptlongint;
    qasSoftState_t *entry;
    
    
    //###FC - DEBUG printk("QoSINUS: qasdev_write called. Offset: %d, count: %d", *ppos, count);
    
    /* Check offset */
    if ( *ppos >= QAS_MAX_NUM_STREAMS * sizeof(qasSoftState_t) )
	{
	    printk("QoSINUS! Trying to write at the end of file.\n");
	    return(-ESPIPE);
	}
    
    /* Copy buffer at current soft state offset */
    ptr = (char *)qasdev_softstate + *ppos;
    ret = copy_from_user(ptr, buf, count);
    
    if ( ret != 0 )
	return(-EFAULT);
    
    //###FC - DEBUG printk(", value: %d.\n", *ptr);  
    
    /* Update offset */
    *ppos = *ppos + count;
    
    /* Check if we've just written a complete record */
    longint = *ppos; //###FC - TMP: kludge to fix a "unresolved symbol __moddi3" problem
    
    if ( ( longint % sizeof(qasSoftState_t) == 0 ) && ( longint != 0 ) )
      {
	    /* Swap bytes */
	    entry = (qasSoftState_t *)((char *)qasdev_softstate + *ppos) - 1;
	    
	    entry->activity = ntohl(entry->activity);
	    ptlongint = (int32_t *)&entry->dscp;

	    *ptlongint = ntohl(*ptlongint); //###FC - TBD: this only works if alignment == 4
	    
	    //###FC - DEBUG printk("QoSINUS: programmed module: stream %d, Dscp %ld\n",  *ppos / sizeof(qasSoftState_t) - 1, *ptlongint);
	}
    
    
    return(count);
}


/*****************************************************************************/
/* Function: qasdev_llseek                                                   */
/* Role    : Called when the pseudo device offset's is changed.              */
/*****************************************************************************/
static loff_t qasdev_llseek(struct file *file, loff_t offset, int whence)
{
  switch ( whence )
    {
    case SEEK_SET:
      if ( ( offset >= QAS_MAX_NUM_STREAMS * sizeof(qasSoftState_t) ) || ( offset < 0 ) )
	return(-EINVAL);
      else
	file->f_pos = offset;
      break;
    case SEEK_CUR:
      if ( ( offset + file->f_pos >= QAS_MAX_NUM_STREAMS * sizeof(qasSoftState_t) ) || ( offset + file->f_pos < 0 ) )
	return(EINVAL);
      else
	file->f_pos += offset;
      break;
    case SEEK_END:
      if ( ( offset >= QAS_MAX_NUM_STREAMS * sizeof(qasSoftState_t) ) || ( offset < 0 ) )
	return(EINVAL);
      else
	file->f_pos -= offset;
      break;
    }
  
  
  /* Done! */
  return(file->f_pos);
}


/*****************************************************************************/
/* Function: init                                                            */
/* Role    : Module entry point.                                             */
/*****************************************************************************/
static int __init init(void)
{
  int ret;
  
  
  /*  Allocate memory where to save the soft state */
  qasdev_softstate = kmalloc(QAS_MAX_NUM_STREAMS * sizeof(qasSoftState_t), GFP_KERNEL);
  
  if ( qasdev_softstate == NULL )
    {
      printk(KERN_WARNING "QoSINUS! can't allocate memory\n");
      return(0);
    }    
  
  memset(qasdev_softstate, 0, QAS_MAX_NUM_STREAMS * sizeof(qasSoftState_t));

  //###FC - XPUT test DO NOT CHECK-IN
  qasdev_softstate[0].dscp = 46;
  //###FC - END XPUT test DO NOT CHECK-IN
  
  /* Initialize pseudo-device */
  ret = register_chrdev(qasdev_major, "qasdev", &qasdev_fops);
  if ( ret < 0 ) 
    {
      printk(KERN_WARNING "QoSINUS! can't get major %d\n", qasdev_major);
      return(ret);
    }
  
  if ( qasdev_major == 0 ) 
      qasdev_major = ret; /* dynamic */

  /* Plug hooks to Netfilter */
  ret = nf_register_hook(&qaspostroute_ops); 
  if ( ret < 0 ) 
      {
	  printk("QoSINUS! can't register local-out hook\n");
	  return ret;
      }
  
  printk("QoSINUS: qasmodule up and running\n");
  
  /* Done! */
  return(0);
}


/*****************************************************************************/
/* Function:                                                                 */
/* Role    :                                                                 */
/*****************************************************************************/
void qasModuleDump()
{
  int i;
  long longint;

  printk("qasmodule dump:\n\n");
    
  for ( i = 0; i < QAS_MAX_NUM_STREAMS ; i++ )
    {
      if ( qasdev_softstate[i].dscp != 0 )
	{
	  longint = qasdev_softstate[i].dscp;
	  printk("%d %d\n", i, longint);
	}
    }
  
  printk("\n\n");
  
  return;
}


/*****************************************************************************/
/* Function: end                                                             */
/* Role    : module unload.                                                  */
/*****************************************************************************/
static void __exit end(void)
{
  //###FC - DEBUG
  qasModuleDump();
  //###FC - DEBUG
  
  /* Unregister pseudo-device */
  unregister_chrdev(qasdev_major, "qasdev");
  
  /* Unregister netfilter hooks */
  nf_unregister_hook(&qaspostroute_ops);
  
  /* Free ressources */
  //###FC - TBD: Free all what needs to be freed
  if ( qasdev_softstate != NULL )
    kfree(qasdev_softstate);
  
  printk("QoSINUS: qasmodule unloaded\n");
}

module_init(init);
module_exit(end);
MODULE_LICENSE("GPL");

