/*****************************************************************************/
/*                                                                           */
/*  qastb.c                                                                  */
/*                                                                           */
/*                                                                           */
/*     Module:                                                               */
/*             qasmodule                                                     */
/*                                                                           */
/*     Description:                                                          */
/*             Token bucket implementation hooked to netfilter.              */
/*                                                                           */
/*     History:                                                              */
/*             11/02/2003 - Fabien Chanussot - Creation                      */
/*                          Based on tbf tc qdisc.                           */
/*                                                                           */
/*****************************************************************************/

#define MODULE 
#define __KERNEL__


/*****************************************************************************/
/* Use                                                                       */
/*****************************************************************************/

#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/init.h>
#include <linux/notifier.h>
#include <linux/netdevice.h>
#include <linux/netfilter.h>
#include <linux/netlink.h>
#include <linux/spinlock.h>
#include <linux/rtnetlink.h>
#include <linux/sysctl.h>
#include <linux/proc_fs.h>
#include <linux/sched.h>

#include "qastbP.h"


/*****************************************************************************/
/* Macros                                                                    */
/*****************************************************************************/

#define L2T(tb,L)   ((tb)->rtab[(L)>>(tb)->resolution])


/*****************************************************************************/
/* Types                                                                     */
/*****************************************************************************/

/*****************************************************************************/
/* Global                                                                    */
/*****************************************************************************/

/* Token bucket */
//qasTokenBucket_t *qasTokenBucket;

/* token buckets table */
qasTokenBucket_t **qasdev_tbs;


/*****************************************************************************/
/* Private functions                                                         */
/*****************************************************************************/

qasTokenBucketEntry_t *qasTokenBucketDequeueEntry(qasTokenBucket_t *theTokenBucket, long *toks);
void qasTokenBucketFlush(qasTokenBucket_t *theTokenBucket);


/*****************************************************************************/
/* Function:                                                                 */
/* Role    :                                                                 */
/*****************************************************************************/
static void qasPulseTimer(unsigned long arg)
{
  qasTokenBucket_t * theTokenBucket;
  
  theTokenBucket = (qasTokenBucket_t *) arg;
  
  /* Dequeue what can be dequeued */
  qasTokenBucketDequeuePackets(theTokenBucket, 1);
}


/*****************************************************************************/
/* Protected functions                                                       */
/*****************************************************************************/

/*****************************************************************************/
/* Public functions                                                          */
/*****************************************************************************/

/*****************************************************************************/
/* Function:                                                                 */
/* Role    :                                                                 */
/*****************************************************************************/
qasTokenBucket_t *qasTokenBucketCreate(long rate, long burst, long backlog)
{
  qasTokenBucket_t *newTb;
  int i;
  unsigned int size;
  unsigned int mtu;
  
  
  /* Allocate new token bucket structure */
  newTb = kmalloc(sizeof(qasTokenBucket_t), GFP_KERNEL); //###FC - TBChecked: GPF_ATOMIC?
  
  if ( newTb == NULL )
    {
      printk(KERN_WARNING "QoSINUS! can't allocate new token bucket\n");
      return(NULL);
    }
  
  /* Initialize the token bucken */
  INIT_LIST_HEAD(&newTb->linkedlist);
  spin_lock_init(&newTb->lock);
  newTb->flushing = 0;
  init_timer(&newTb->timer);
  newTb->timer.function = qasPulseTimer;
  newTb->timer.data = (unsigned long)newTb;
  newTb->timerperiod = 1; //###FC - TBD: do that based on the rate
  newTb->len = 0;
  newTb->rate = rate;
  newTb->burst = 1000000 * ((double)burst / newTb->rate);
  printk("QoSINUS: burst = %ld\n", newTb->burst); //###FC - DEBUG
  newTb->backlog = backlog;
  PSCHED_GET_TIME(newTb->timeref);
  newTb->tokens = 0;
  
  mtu = 2047; //###FC - TBD: do we need to control that?
  
  newTb->resolution = 0;
  while ( ( mtu >> newTb->resolution ) > 255 )
    newTb->resolution++;
  
  printk("QoSINUS: rate table:\n");
  for ( i = 0; i < 256; i++ )
    {
	size = (i << newTb->resolution);
	newTb->rtab[i] = (u32) (1000000 * ((double)size / newTb->rate));
	printk("tab[%d] = %ld\n", i, newTb->rtab[i]);
    }
  
  printk("QoSINUS: created token bucket: rate %ld burst %ld backlog %ld\n", newTb->rate, newTb->burst, newTb->backlog);
  
  
  /* Done! */
  return(newTb);
  //###FC -TBD: release everything
}


/*****************************************************************************/
/* Function:                                                                 */
/* Role    :                                                                 */
/*****************************************************************************/
void qasTokenBucketChange(qasTokenBucket_t *tb, long rate, long burst)
{
  //###FC - TBD
}


/*****************************************************************************/
/* Function:                                                                 */
/* Role    :                                                                 */
/*****************************************************************************/
void qasTokenBucketDelete(qasTokenBucket_t *tb)
{
  spin_lock_bh(&tb->lock);
  tb->flushing = 1;
  spin_unlock_bh(&tb->lock);
  del_timer_sync(&tb->timer);
  qasTokenBucketFlush(tb);
}


/*****************************************************************************/
/* Function:                                                                 */
/* Role    :                                                                 */
/*****************************************************************************/
int qasTokenBucketEnqueuePacket(struct sk_buff *skb, struct nf_info *info, void *data)
{
  qasTokenBucket_t *theTokenBucket;
  qasTokenBucketEntry_t *newEntry, *anEntry;
  int ret = 0;
  
  
  printk("# on %d\n", smp_processor_id());
  
  /* Find the token bucket defined for the stream the packet belongs to */
  //$$$
  theTokenBucket = (qasTokenBucket_t *)data;
  //$$$
  //$$$ theTokenBucket = (qasTokenBucket_t *)qasdev_tbs[skb->nfmark]; //###FC - TBChecked: make sure that no lock is needed
  
  /* Create new entry */
  newEntry = kmalloc(sizeof(qasTokenBucketEntry_t), GFP_ATOMIC);
  
  if ( newEntry == NULL )
    {
      printk(KERN_WARNING "QoSINUS! can't allocate new token bucket entry\n");
      return(-ENOMEM);
    }
  
  newEntry->skb = skb;
  newEntry->info = info;
  
  /* Queue new packet */  
  spin_lock(&theTokenBucket->lock);
  list_add_tail(&newEntry->list, &theTokenBucket->linkedlist);
  theTokenBucket->len += skb->len; //###FC - TBChecked: so far, we only take the payload size into account.
  /////spin_unlock(&theTokenBucket->lock); 
  
  /* Dequeue the conforming flow */
  qasTokenBucketDequeuePackets(theTokenBucket, 0);
  
  /* Check if the new packet should be dropped */
  /////spin_lock(&theTokenBucket->lock);
  if ( theTokenBucket->len > theTokenBucket->backlog )
    { //Drop it!
      anEntry = (qasTokenBucketEntry_t *)theTokenBucket->linkedlist.prev; //###FC - TBChecked: heu, c'est pas justement newEntry???
      
      theTokenBucket->len -= skb->len;
      
      /////spin_unlock(&theTokenBucket->lock);
      
      nf_reinject(anEntry->skb, anEntry->info, NF_DROP);
      printk("QoSINUS: tb dropped packet\n");
      
      /* Remove the entry from the queue */
      /////spin_lock(&theTokenBucket->lock);
      list_del(&anEntry->list);
      /////spin_unlock(&theTokenBucket->lock);
      kfree(anEntry);
    }
  /////else
  spin_unlock(&theTokenBucket->lock);
  
  
  /* Done! */
  return(ret);
}


/*****************************************************************************/
/* Function:                                                                 */
/* Role    :                                                                 */
/*****************************************************************************/
qasTokenBucketEntry_t *qasTokenBucketDequeueEntry(qasTokenBucket_t *theTokenBucket, long *toks)
{
  long deltatoks;
  struct list_head *i;
  qasTokenBucketEntry_t *anEntry;
  
  
  /////spin_lock(&theTokenBucket->lock);
  
  i = theTokenBucket->linkedlist.next;
  
  if ( i != &theTokenBucket->linkedlist )
    {
      anEntry = (qasTokenBucketEntry_t *)i;
      deltatoks =  L2T(theTokenBucket, anEntry->skb->len);
      
      if ( *toks >= deltatoks )
	{
	  *toks -= deltatoks;
	  //###FC - TBD theTokenBucket->tokens = *toks;
	  theTokenBucket->len -=  anEntry->skb->len;
	  list_del(&anEntry->list);
	}
      else
	anEntry = NULL;
      
      /////spin_unlock(&theTokenBucket->lock);
      
      return(anEntry);
    }

  /////spin_unlock(&theTokenBucket->lock);
  
  
  return(NULL);
}


/*****************************************************************************/
/* Function:                                                                 */
/* Role    :                                                                 */
/*****************************************************************************/
void qasTokenBucketDequeuePackets(qasTokenBucket_t *theTokenBucket, int pulse)
{
  qasTokenBucketEntry_t *anEntry;
  int haSacadoAlgo = 0;
  long toks;
  psched_time_t now;
  int leftover = 0;
  
  
  /* Evaluate available token number */
  PSCHED_GET_TIME(now);
  
  /////spin_lock(&theTokenBucket->lock);
  toks = PSCHED_TDIFF_SAFE(now, theTokenBucket->timeref, theTokenBucket->burst, 0);
  /////spin_unlock(&theTokenBucket->lock);
  
  //###FC - DEBUG toks += theTokenBucket->tokens;
  
  /* Limit the burst */
  if ( toks > (long)theTokenBucket->burst )
    toks = theTokenBucket->burst;
  
  /* Reinject conform flow */
  while ( ( toks >= 0 ) && ( anEntry = qasTokenBucketDequeueEntry(theTokenBucket, &toks) ) )
    {
      //###FC - DEBUG printk("QoSINUS: trying to reinject packet, toks=%ld, sub=%ld\n", toks, L2T(theTokenBucket, anEntry->skb->len)); //###FC - DEBUG
      nf_reinject(anEntry->skb, anEntry->info, NF_ACCEPT);
      kfree(anEntry);
      
      if ( !haSacadoAlgo )
	{  // Update timer only once 
	  /////spin_lock(&theTokenBucket->lock);
	  theTokenBucket->timeref = now;
	  /////spin_unlock(&theTokenBucket->lock);
	}
      
      haSacadoAlgo++;
    }
  
  /* Return the amount of packet bytes left */
  //###FC - TBChecked spin_lock(&theTokenBucket->lock);
  leftover = theTokenBucket->len;
  //###FC - TBChecked spin_unlock(&theTokenBucket->lock);
  
  /* Arm timer */
  if ( leftover > 0 )
    {
      /////spin_lock(&theTokenBucket->lock);
      if ( !theTokenBucket->flushing )
	mod_timer(&theTokenBucket->timer, jiffies + theTokenBucket->timerperiod);
      /////spin_unlock(&theTokenBucket->lock);
    }
  
  
  /* Done!*/
  return;
}


/*****************************************************************************/
/* Function:                                                                 */
/* Role    :                                                                 */
/*****************************************************************************/
void qasTokenBucketFlush(qasTokenBucket_t *theTokenBucket)
{
  struct list_head *i;
  qasTokenBucketEntry_t *anEntry;
  
  
  /* Reinject all the packets left in the token bucket */
  spin_lock_bh(&theTokenBucket->lock);
  
  i = theTokenBucket->linkedlist.next;
  
  while ( i != &theTokenBucket->linkedlist )
    {
      anEntry = (qasTokenBucketEntry_t *)i;
      
      list_del(&anEntry->list);
      
      spin_unlock_bh(&theTokenBucket->lock);

      nf_reinject(anEntry->skb, anEntry->info, NF_ACCEPT);
      kfree(anEntry);
      
      spin_lock_bh(&theTokenBucket->lock);
    }

  spin_unlock_bh(&theTokenBucket->lock);
}
