///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//  qos.c                                                                    //
//                                                                           //
//                                                                           //
//     Module:                                                               //
//             qos.so library                                                //
//                                                                           //
//     Description:                                                          //
//             QoS class methods that let applications specify their         //
//             requirements and invoke QoS sessions.                         //
//                                                                           //
//     History:                                                              //
//             30/08/2002 - Fabien Chanussot - Creation                      //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Use                                                                       //
///////////////////////////////////////////////////////////////////////////////

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <xercesc/framework/MemBufFormatTarget.hpp>
#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/dom/DOM.hpp>

#include <xercesc/parsers/XercesDOMParser.hpp>
#include <xercesc/framework/MemBufInputSource.hpp> 
#include "DOMTreeErrorReporter.hpp"
#include "DOMPrintFilter.hpp"
#include "DOMPrintErrorHandler.hpp"
#include "activePacket.h"

#include "qos.hpp"


///////////////////////////////////////////////////////////////////////////////
// Configuration data                                                        //
///////////////////////////////////////////////////////////////////////////////

char *defaultSkeleton="/.QoSINUS/skeleton.xml";


///////////////////////////////////////////////////////////////////////////////
// Macros                                                                    //
///////////////////////////////////////////////////////////////////////////////

#define CU_MASK 0x03


///////////////////////////////////////////////////////////////////////////////
// Types                                                                     //
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// Global variables                                                          //
///////////////////////////////////////////////////////////////////////////////

static XercesDOMParser::ValSchemes gValScheme = XercesDOMParser::Val_Auto;
static bool gDoNamespaces = false;
static bool gDoSchema = false;
static bool gSchemaFullChecking = false;
static bool gDoCreate = true;
static const XMLCh* gOutputEncoding = 0;
static const XMLCh* gMyEOLSequence = 0;
static bool gSplitCdataSections = true;
static bool gDiscardDefaultContent = true;
static bool gFormatPrettyPrint = false;


//////////////////////////////////////////////////////////////////////////////
// Private functions                                                         //
///////////////////////////////////////////////////////////////////////////////

// ---------------------------------------------------------------------------
//  This is a simple class that lets us do easy (though not terribly efficient)
//  trancoding of char* data to XMLCh data.
// ---------------------------------------------------------------------------
class XStr
{
public :
    // -----------------------------------------------------------------------
    //  Constructors and Destructor
    // -----------------------------------------------------------------------
    XStr(const char* const toTranscode)
    {
        // Call the private transcoding method
        fUnicodeForm = XMLString::transcode(toTranscode);
    }

    ~XStr()
    {
        delete [] fUnicodeForm;
    }


    // -----------------------------------------------------------------------
    //  Getter methods
    // -----------------------------------------------------------------------
    const XMLCh* unicodeForm() const
    {
        return fUnicodeForm;
    }

private :
    // -----------------------------------------------------------------------
    //  Private data members
    //
    //  fUnicodeForm
    //      This is the Unicode XMLCh format of the string.
    // -----------------------------------------------------------------------
    XMLCh*   fUnicodeForm;
};

#define X(str) XStr(str).unicodeForm()


///////////////////////////////////////////////////////////////////////////////
// Function: getBaseSLS                                                      //
// Role    : find the xml skeleton based on delay, rate and loss             //
///////////////////////////////////////////////////////////////////////////////
int getBaseSLS(char *filename, int delay, int rate, int loss)
{
  //  QoS_M_errormsg("getBaseSLS not implemented yet...");
  
  //Just on value skeleton so far
  strcpy(filename, getenv("HOME"));
  strcat(filename, defaultSkeleton);
  
  
  /* Done! */
  return(0);
}


///////////////////////////////////////////////////////////////////////////////
// Function: parseXmlFileMem                                                 //
// Role    : create the XML DOM object                                       //
///////////////////////////////////////////////////////////////////////////////
int parseXmlFileMem(XercesDOMParser *parser, DOMDocument **slsIn, char *filename, InputSource &source)
{
  //###TBD: use XML Schemas
  //  Create our parser, then attach an error handler to the parser.
  //  The parser will call back to methods of the ErrorHandler if it
  //  discovers errors during the course of parsing the XML document.
  parser->setValidationScheme(gValScheme);
  parser->setDoNamespaces(gDoNamespaces);
  parser->setDoSchema(gDoSchema);
  parser->setValidationSchemaFullChecking(gSchemaFullChecking);
  parser->setCreateEntityReferenceNodes(gDoCreate);
  
  DOMTreeErrorReporter *errReporter = new DOMTreeErrorReporter();
  parser->setErrorHandler(errReporter);
  
  //  Parse the XML file, catching any XML exceptions that might propogate
  //  out of it.
  bool errorsOccured = false;
  try
    {
      if ( filename != NULL )
	parser->parse(filename);
      else
	parser->parse(source);
    }
  
  catch (const XMLException& e)
    { //###TBD: use my macros
      cerr << "An error occurred during parsing\n"
	   << endl;
      errorsOccured = true;
    }
  
  catch (const DOMException& e)
    { //###TBD: use my macros
      cerr << "A DOM error occurred during parsing\n   DOMException code: "
	   << e.code << endl;
      errorsOccured = true;
    }
  
  catch (...)
    { //###TBD: use my macros
      cerr << "An error occurred during parsing\n " << endl;
      errorsOccured = true;
    }
  
  // Get the DOM representation
  *slsIn = parser->getDocument();
  
  //  Clean up the error handler. The parser does not adopt handlers
  //  since they could be many objects or one object installed for multiple
  //  handlers.
  delete errReporter;
  
  // If the parse was successful, output the document data from the DOM tree
  if ( errorsOccured )
    return(-2);
  
  
  /* Done! */
  return(0);
}


///////////////////////////////////////////////////////////////////////////////
// Function: appendValue                                                     //
// Role    : append a qualitative or quantitative value element              //
///////////////////////////////////////////////////////////////////////////////
int appendValue(DOMDocument *slsIn, DOMElement *parentElem, char *childName, int childValue)
{
  DOMElement* childElem = slsIn->createElement(X(childName));
  char quantitative[255];
  
  
  switch (childValue)
    {
    case QOS_HIGH:
      childElem->setAttribute(X("value"), X("high"));
      break;
    case QOS_MEDIUM:
      childElem->setAttribute(X("value"), X("medium"));
      break;
    case QOS_LOW:
      childElem->setAttribute(X("value"), X("low"));
      break;
    case QOS_ANY:
      //Do nothing
      break;
    default:
      sprintf(quantitative, "%d", childValue);
      childElem->setAttribute(X("value"), X(quantitative));
      break;
    }
  
  parentElem->appendChild(childElem);
  
  
  /* Done! */
  return(0);
}


///////////////////////////////////////////////////////////////////////////////
// Function: findChildByName                                                 //
// Role    : Find the first child with a given name                          //
///////////////////////////////////////////////////////////////////////////////
DOMElement *findChildByName(DOMNode* parent, char *name)
{
  DOMNode *child = parent->getFirstChild();
  DOMElement *newElem = 0;
  bool trouve = false;
  
  while ( child != 0 && !trouve )
    {
      if ( !XMLString::compareString(child->getNodeName(), X(name)) )
	trouve = true;
      else
	child = child->getNextSibling();
    }
  
  if ( child != 0 )
    if ( child->getNodeType() == 1 ) // ELEMENT_NODE = 1
      newElem = (DOMElement *)child;
  
  
  return(newElem);
}


///////////////////////////////////////////////////////////////////////////////
// Public functions/methods                                                  //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Method  : Set                                                             //
// Role    : simplified request of QoS                                       //
///////////////////////////////////////////////////////////////////////////////
int QoS :: Set(char *ipdest, char *portdest, int delay, int rate, int loss, const char *ipstreamdest, char *QoSCode)
{
  //###TBD: manage the flow type
  //###TBD: en general error recovery clean dans ce fichier
  DOMDocument *slsIn;
  DOMDocument *slsOut;
  int iret;  char filename[PATH_MAX+NAME_MAX];
  
  
  /* Build a SLS based on delay, rate and loss values */
  // Find the xml skeleton that matches delay, rate and loss
  iret = getBaseSLS(filename, delay, rate, loss);
  
  if ( iret != 0 )
    return(iret);
  
  // Initialize the XML system
  try
    {
      XMLPlatformUtils::Initialize();
    }
  
  catch(const XMLException &toCatch)
    { //###TBD : Use my macros
      cerr << "Error during Xerces-c Initialization.\n"
	   << endl;
      return(-1);
    }
  
  // Create XML DOM object
  XercesDOMParser *parser = new XercesDOMParser; //Has to be here
  MemBufInputSource dummyMemSource(NULL, 0, (char *) NULL, false);
  iret = parseXmlFileMem(parser, &slsIn, filename, dummyMemSource);
  
  if ( iret != 0 )
    {
      XMLPlatformUtils::Terminate();
      return(iret);
    }
  
  // Add missing branches
  DOMElement* rootElem = slsIn->getDocumentElement();
  
  rootElem->setAttribute(X("type"), X("QOS_REQ"));
  
  DOMElement* flowidElem = slsIn->createElement(X("flowid"));
  rootElem->appendChild(flowidElem);
  
  flowidElem->setAttribute(X("destination"), X(ipstreamdest));
  
  DOMElement* guaranteesElem = slsIn->createElement(X("guarantees"));
  rootElem->appendChild(guaranteesElem);
  
  appendValue(slsIn, guaranteesElem, "delay", delay);
  appendValue(slsIn, guaranteesElem, "loss", loss);
  appendValue(slsIn, guaranteesElem, "rate", rate);
  
  /* Call the negociate method */
  iret = this->Request(ipdest, portdest, slsIn, &slsOut);
  
  if ( iret != 0 )
    {
      XMLPlatformUtils::Terminate();
      printf("###FC - Error: request returned %d\n", iret); //###FC - DEBUG      
      return(iret);
    }
  
  /* Retrieve the DSCP to be used */
  rootElem = slsOut->getDocumentElement();
  DOMElement *child = findChildByName(rootElem, "flowid");
  
  if ( child == 0 )
    {
      printf("###FC - Error: no flowid child\n"); //###FC - DEBUG
      XMLPlatformUtils::Terminate();
      return(iret);
    }
  
  const XMLCh *dscpStr;
  dscpStr = child->getAttribute(X("dscp"));
  if ( *dscpStr == 0 )
    {
      printf("###FC - Error: no dscp attribute\n"); //###FC - DEBUG
      XMLPlatformUtils::Terminate();
      return(iret);
    }
  
  *QoSCode = (char) XMLString::parseInt(dscpStr); //###TBD: do some checking here (too big of a value)?
  //###FC - DEBUG printf("###DSCP read: %d\n", *QoSCode);
  
  /* Free resources */
  // Delete the parser itself.  Must be done prior to calling Terminate, below.
  delete parser;
  //###TBD: more?
  
  // And call the termination method
  XMLPlatformUtils::Terminate();
  
  
  /* Done! */
  return(0);
}


///////////////////////////////////////////////////////////////////////////////
// Method  : Request                                                         //
// Role    : request QoS from the system                                     //
///////////////////////////////////////////////////////////////////////////////
int QoS :: Request(char *ipdest, char *portdest, DOMDocument *SLSin, DOMDocument **SLSout)
{
  int s;
  int iret;
  struct protoent *ppe;
  struct sockaddr_in sin;
  struct servent *pse;
  struct hostent *phe;
  activeSocketHandle_t *hActSock;
  char *tmpBuff, *newBuff;
  size_t buffSize, tmpSize;
  MemBufFormatTarget *myFormTarget = new MemBufFormatTarget();
  
  
  /* Create/connect/etc a socket */
  if ( (ppe = getprotobyname("tcp")) == 0 )
    {
      QoS_M_error(0);
      return(-1);
    }
  
  s = socket(PF_INET, SOCK_STREAM, ppe->p_proto);
  
  if ( s < 0 )
    {
      QoS_M_error(s);
      return(s);
    }
  
  memset(&sin, 0, sizeof(sin));
  sin.sin_family = AF_INET;
  
  if ( (pse = getservbyname(portdest, "tcp")) )
    sin.sin_port = pse->s_port;
  else if ( ( sin.sin_port = htons((u_short)atoi(portdest)) ) == 0 )
    {
      QoS_M_error(0);
      return(-1);
    }
  
  if ( (phe = gethostbyname(ipdest)) )
    memcpy((char *)&sin.sin_addr, phe->h_addr, phe->h_length);
  else if ( ( sin.sin_addr.s_addr = inet_addr(ipdest) ) == INADDR_NONE )
    {
      QoS_M_error(INADDR_NONE);
      return(-1);
    }
  
  iret = connect(s, (struct sockaddr *)&sin, sizeof(sin));
  
  if ( iret < 0 )
    {
      QoS_M_error(iret);
      return(iret);
    }
  
  /* "Activate" the socket */
  hActSock = activeSocket(&s, NATIVE_SOCKET, "QaS", "0.0.0.0", ipdest); //The last field should be filled in when using TCP
  
  if ( hActSock == NULL )
    {
      QoS_M_error(0);
      return(-1);
    }
  
  /* Send the QoS_REQ message */
  try
    {
      // get a serializer, an instance of DOMWriter
      XMLCh tempStr[100];
      XMLString::transcode("LS", tempStr, 99);
      DOMImplementation *impl = DOMImplementationRegistry::getDOMImplementation(tempStr);
      DOMWriter *theSerializer = ((DOMImplementationLS*)impl)->createDOMWriter();
      
      // set user specified end of line sequence and output encoding
      theSerializer->setNewLine(gMyEOLSequence);
      theSerializer->setEncoding(gOutputEncoding);
      
      // plug in user's own error handler
      DOMErrorHandler *myErrorHandler = new DOMPrintErrorHandler();
      theSerializer->setErrorHandler(myErrorHandler);
      
      // set feature if the serializer supports the feature/mode
      if (theSerializer->canSetFeature(XMLUni::fgDOMWRTSplitCdataSections, gSplitCdataSections))
	theSerializer->setFeature(XMLUni::fgDOMWRTSplitCdataSections, gSplitCdataSections);
      
      if (theSerializer->canSetFeature(XMLUni::fgDOMWRTDiscardDefaultContent, gDiscardDefaultContent))
	theSerializer->setFeature(XMLUni::fgDOMWRTDiscardDefaultContent, gDiscardDefaultContent);
      
      if (theSerializer->canSetFeature(XMLUni::fgDOMWRTFormatPrettyPrint, gFormatPrettyPrint))
	theSerializer->setFeature(XMLUni::fgDOMWRTFormatPrettyPrint, gFormatPrettyPrint);
      
      // Plug in a format target to receive the resultant
      // XML stream from the serializer.
      //
      // StdOutFormatTarget prints the resultant XML stream
      // to stdout once it receives any thing from the serializer.
      //
      
      // do the serialization through DOMWriter::writeNode();
      theSerializer->writeNode(myFormTarget, *SLSin);
      
      // Build the active paquet and send it.
      //###FC - DEBUG printf("%s\n", myFormTarget->getRawBuffer());
      iret = activeWrite(hActSock, (char *) myFormTarget->getRawBuffer(), myFormTarget->getLen());
      
      if ( iret != 0 )
	{ //###TBD
	  QoS_M_error(0);
	  return(-1);
	}
      
      delete theSerializer;
      
      // Filter, formatTarget and error handler
      // are NOT owned by the serializer.
      delete myFormTarget;
      delete myErrorHandler;
    }
  catch (XMLException& e)
    {
      cerr << "An error occurred during creation of output transcoder."
	   << endl;
      
      QoS_M_error(0);
      return(-1);
    }
  
  /* Read QoS_ACK message */
  buffSize = myFormTarget->getLen() * 3 / 2; //To start with something...
  
  tmpSize = buffSize;
  tmpBuff = (char *)malloc(buffSize);
  
  if ( tmpBuff == NULL )
    {
      //###TBD: Error handling
      QoS_M_error(0);
      return(-1);
    }
  
  iret = activeRead(hActSock, tmpBuff, &tmpSize);
  
  if ( iret != 0 )
    {
      //###TBD: Error handling
      QoS_M_error(iret);
      return(iret);
    }
  
  if ( tmpSize > buffSize )
    { // Read the payload leftover
      newBuff = (char *)realloc(tmpBuff, tmpSize);
      
      if ( newBuff == NULL )
	  {
	  free(tmpBuff);
	  //###TBD: Error handling
	  QoS_M_error(0);
	  return(-1);
	}
      
      tmpBuff = newBuff;
      
      tmpSize -= buffSize;
      iret = activeRead(hActSock, tmpBuff+buffSize, &tmpSize);
      
      if ( iret != 0 )
	{
	  //###TBD: Error handling
	  QoS_M_error(0);
	  return(iret);
	}
      
      tmpSize += buffSize;
    }
  
  //###TBD: use inputsource to parse in memory and create SLSout
  XercesDOMParser *parser = new XercesDOMParser;
  MemBufInputSource dummyMemSource((XMLByte *)tmpBuff, (unsigned int)tmpSize, (char *) NULL, false);
  iret = parseXmlFileMem(parser, SLSout, NULL, dummyMemSource);
  
  // Free resources
  free(tmpBuff);
  activeClose(hActSock);

  //###TBD: more resources to free up (dummyMemSource, ...)?
  
  
  
  /* Done! */
  return(0);
}


///////////////////////////////////////////////////////////////////////////////
// Method  : Invoke                                                          //
// Role    : program socket to mark DSCP field                               //
///////////////////////////////////////////////////////////////////////////////
int QoS :: Invoke(int sock, char QoSCode)
{
  socklen_t len;
  char ctemp;
  
  
  /* The two last bits of the DSCP are reserved for future use: check that QoSCode is not too big... */
  if ( QoSCode > 63 )
    {
      QoS_M_error(QoSCode);
      return(-1);
    }
  
  /* Read the CU bits of the TOS field. */
  len = 1;
  if ( getsockopt(sock, IPPROTO_IP, IP_TOS, (void *)&(ctemp), &len) == -1 )
    {
      QoS_M_errno(errno);
      return(-1);
    }
  
  ctemp &= CU_MASK;
  
  /* Define the new TOS field */
  if ( QoSCode != -1 )
    QoSCode = (QoSCode << 2) | ctemp;
  else
    QoSCode = ctemp; //###FC - TBD: if received -1: restore DSCP previous programming
  
  /* Set the socket's IP_TOS option */
  if ( setsockopt(sock, IPPROTO_IP, IP_TOS, (void *)&QoSCode, 1) == -1 )
    {
      QoS_M_errno(errno);
      return(-1);
    }
  
  
  /* Done! */
  return(0);
}


///////////////////////////////////////////////////////////////////////////////
// Method  : Release                                                         //
// Role    : release the DSCP assigned to the session                        //
///////////////////////////////////////////////////////////////////////////////
int QoS :: Release(char *ipdest, char *portdest, char QoSCode)
{
    int s;
    int iret;
    struct protoent *ppe;
    struct sockaddr_in sin;
    struct servent *pse;
    struct hostent *phe;
    activeSocketHandle_t *hActSock;
    MemBufFormatTarget *myFormTarget = new MemBufFormatTarget();
    DOMDocument *slsIn;
    char filename[PATH_MAX+NAME_MAX];
    char dscpStr[255];
    char *tmpBuff, *newBuff;
    size_t buffSize, tmpSize;
    
    
    //Use the skeleton as XML file basis
    strcpy(filename, getenv("HOME"));
    strcat(filename, defaultSkeleton);

    // Initialize the XML system
    try
	{
	    XMLPlatformUtils::Initialize();
	}
    
    catch(const XMLException &toCatch)
	{ //###TBD : Use my macros
	    cerr << "Error during Xerces-c Initialization.\n"
		 << endl;
	    return(-1);
	}
    
    // Create XML DOM object
    XercesDOMParser *parser = new XercesDOMParser; //Has to be here
    MemBufInputSource dummyMemSource(NULL, 0, (char *) NULL, false);
    iret = parseXmlFileMem(parser, &slsIn, filename, dummyMemSource);
    
    if ( iret != 0 )
	{
	    XMLPlatformUtils::Terminate();
	    return(iret);
	}
    
    // Add missing branches
    DOMElement* rootElem = slsIn->getDocumentElement();
    
    rootElem->setAttribute(X("type"), X("QOS_RST"));
    
    DOMElement* flowidElem = slsIn->createElement(X("flowid"));
    rootElem->appendChild(flowidElem);
    
    sprintf(dscpStr, "%d", (int)QoSCode);
    flowidElem->setAttribute(X("dscp"), X(dscpStr));
    
    // Send the RST message
    //###TBD - par la meme occasion, ajouter le type de message dans le SLS...
    /* Create/connect/etc a socket */
    if ( (ppe = getprotobyname("tcp")) == 0 )
	{
	    QoS_M_error(0);
	    return(-1);
	}
    
    s = socket(PF_INET, SOCK_STREAM, ppe->p_proto);
    
    if ( s < 0 )
	{
	    QoS_M_error(s);
	    return(s);
	}
    
    memset(&sin, 0, sizeof(sin));
    sin.sin_family = AF_INET;
    
    if ( (pse = getservbyname(portdest, "tcp")) )
	sin.sin_port = pse->s_port;
    else if ( ( sin.sin_port = htons((u_short)atoi(portdest)) ) == 0 )
	{
	    QoS_M_error(0);
	    return(-1);
	}
    
    if ( (phe = gethostbyname(ipdest)) )
	memcpy((char *)&sin.sin_addr, phe->h_addr, phe->h_length);
    else if ( ( sin.sin_addr.s_addr = inet_addr(ipdest) ) == INADDR_NONE )
	{
	    QoS_M_error(INADDR_NONE);
	    return(-1);
	}
    
    iret = connect(s, (struct sockaddr *)&sin, sizeof(sin));
    
    if ( iret < 0 )
	{
	    QoS_M_error(iret);
	    return(iret);
	}
    
    /* "Activate" the socket */
    hActSock = activeSocket(&s, NATIVE_SOCKET, "QaS", "0.0.0.0", ipdest); //The last field should be filled in when using TCP
    
    if ( hActSock == NULL )
	{
	    QoS_M_error(0);
	    return(-1);
	}
    
    /* Send the QoS_RST message */
    try
	{
	    // get a serializer, an instance of DOMWriter
	    XMLCh tempStr[100];
	    XMLString::transcode("LS", tempStr, 99);
	    DOMImplementation *impl = DOMImplementationRegistry::getDOMImplementation(tempStr);
	    DOMWriter *theSerializer = ((DOMImplementationLS*)impl)->createDOMWriter();
	    
	    // set user specified end of line sequence and output encoding
	    theSerializer->setNewLine(gMyEOLSequence);
	    theSerializer->setEncoding(gOutputEncoding);
	    
	    // plug in user's own error handler
	    DOMErrorHandler *myErrorHandler = new DOMPrintErrorHandler();
	    theSerializer->setErrorHandler(myErrorHandler);
	    
	    // set feature if the serializer supports the feature/mode
	    if (theSerializer->canSetFeature(XMLUni::fgDOMWRTSplitCdataSections, gSplitCdataSections))
		theSerializer->setFeature(XMLUni::fgDOMWRTSplitCdataSections, gSplitCdataSections);
	    
	    if (theSerializer->canSetFeature(XMLUni::fgDOMWRTDiscardDefaultContent, gDiscardDefaultContent))
		theSerializer->setFeature(XMLUni::fgDOMWRTDiscardDefaultContent, gDiscardDefaultContent);
	    
	    if (theSerializer->canSetFeature(XMLUni::fgDOMWRTFormatPrettyPrint, gFormatPrettyPrint))
		theSerializer->setFeature(XMLUni::fgDOMWRTFormatPrettyPrint, gFormatPrettyPrint);
	    
	    // Plug in a format target to receive the resultant
	    // XML stream from the serializer.
	    //
	    // StdOutFormatTarget prints the resultant XML stream
	    // to stdout once it receives any thing from the serializer.
	    //
	    
	    // do the serialization through DOMWriter::writeNode();
	    theSerializer->writeNode(myFormTarget, *slsIn);
	    
	    // Build the active paquet and send it.
	    //###FC - DEBUG printf("%s\n", myFormTarget->getRawBuffer());
	    iret = activeWrite(hActSock, (char *) myFormTarget->getRawBuffer(), myFormTarget->getLen());
	    
	    if ( iret != 0 )
		{ //###TBD
		    QoS_M_error(0);
		    return(-1);
		}
	    
	    delete theSerializer;
	    
	    // Filter, formatTarget and error handler
	    // are NOT owned by the serializer.
	    delete myFormTarget;
	    delete myErrorHandler;
	}
    catch (XMLException& e)
	{
	    cerr << "An error occurred during creation of output transcoder."
		 << endl;
	    
	    QoS_M_error(0);
	    return(-1);
	}

    // Read ACK message
    buffSize = myFormTarget->getLen() * 3 / 2; //To start with something...
  
    tmpSize = buffSize;
    tmpBuff = (char *)malloc(buffSize);
    
    if ( tmpBuff == NULL )
	{
	    //###TBD: Error handling
	    QoS_M_error(0);
	    return(-1);
	}
    
    iret = activeRead(hActSock, tmpBuff, &tmpSize);
    
    if ( iret != 0 )
	{
	    //###TBD: Error handling
	    QoS_M_error(iret);
	    return(iret);
	}
  
    if ( tmpSize > buffSize )
	{ // Read the payload leftover
	    newBuff = (char *)realloc(tmpBuff, tmpSize);
	    
	    if ( newBuff == NULL )
		{
		    free(tmpBuff);
		    //###TBD: Error handling
		    QoS_M_error(0);
		    return(-1);
		}
	    
	    tmpBuff = newBuff;
	    
	    tmpSize -= buffSize;
	    iret = activeRead(hActSock, tmpBuff+buffSize, &tmpSize);
	    
	    if ( iret != 0 )
		{
		    //###TBD: Error handling
		    QoS_M_error(0);
		    return(iret);
		}
	    
	    tmpSize += buffSize;
	}
    
    // Free resources
    free(tmpBuff);
    activeClose(hActSock);
    
    // Delete the parser itself.  Must be done prior to calling Terminate, below.
    delete parser;
    //###TBD: more?
    
    // And call the termination method
    XMLPlatformUtils::Terminate();
    
    
    /* Done! */
    return(0);
}

